"""
MIT License

Copyright (c) 2024 Tom Quirk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import random
import base64
from typing import Dict, List


def get_id_from_urn(urn: str):
    """
    Return the ID of a given Linkedin URN.

    Example: urn:li:fs_miniProfile:<id>
    """
    return urn.split(":")[3]


def get_urn_from_raw_update(raw_string: str) -> str:
    """
    Return the URN of a raw group update

    Example: urn:li:fs_miniProfile:<id>
    Example: urn:li:fs_updateV2:(<urn>,GROUP_FEED,EMPTY,DEFAULT,false)
    """
    return raw_string.split("(")[1].split(",")[0]


def get_update_author_name(d_included: Dict) -> str:
    """Parse a dict and returns, if present, the post author name

    :param d_included: a dict, as returned by res.json().get("included", {})
    :type d_raw: dict

    :return: Author name
    :rtype: str
    """
    try:
        return d_included["actor"]["name"]["text"]
    except KeyError:
        return ""
    except TypeError:
        return "None"


def get_update_old(d_included: Dict) -> str:
    """Parse a dict and returns, if present, the post old string

    :param d_included: a dict, as returned by res.json().get("included", {})
    :type d_raw: dict

    :return: Post old string. Example: '2 mo'
    :rtype: str
    """
    try:
        return d_included["actor"]["subDescription"]["text"]
    except KeyError:
        return ""
    except TypeError:
        return "None"


def get_update_content(d_included: Dict, base_url: str) -> str:
    """Parse a dict and returns, if present, the post content

    :param d_included: a dict, as returned by res.json().get("included", {})
    :type d_raw: dict
    :param base_url: site URL
    :type d_raw: str

    :return: Post content
    :rtype: str
    """
    try:
        return d_included["commentary"]["text"]["text"]
    except KeyError:
        return ""
    except TypeError:
        # Let's see if its a reshared post...
        try:
            # TODO: call Linkedin API to fetch that particular post and extract content
            urn = get_urn_from_raw_update(d_included["*resharedUpdate"])
            return f"{base_url}/feed/update/{urn}"
        except KeyError:
            return "IMAGE"
        except TypeError:
            return "None"


def get_update_author_profile(d_included: Dict, base_url: str) -> str:
    """Parse a dict and returns, if present, the URL corresponding the profile

    :param d_included: a dict, as returned by res.json().get("included", {})
    :type d_raw: dict
    :param base_url: site URL
    :type d_raw: str

    :return: URL with either company or member profile
    :rtype: str
    """
    urn: str = ""
    try:
        urn = d_included["actor"]["urn"]
    except KeyError:
        return ""
    except TypeError:
        return "None"
    else:
        urn_id = urn.split(":")[-1]
        if "company" in urn:
            return f"{base_url}/company/{urn_id}"
        elif "member" in urn:
            return f"{base_url}/in/{urn_id}"
    return urn


def get_update_url(d_included: Dict, base_url: str) -> str:
    """Parse a dict and returns, if present, the post URL

    :param d_included: a dict, as returned by res.json().get("included", {})
    :type d_raw: dict
    :param base_url: site URL
    :type d_raw: str

    :return: post url
    :rtype: str
    """
    try:
        urn = d_included["updateMetadata"]["urn"]
    except KeyError:
        return ""
    except TypeError:
        return "None"
    else:
        return f"{base_url}/feed/update/{urn}"


def append_update_post_field_to_posts_list(
    d_included: Dict, l_posts: List, post_key: str, post_value: str
) -> List[Dict]:
    """Parse a dict and returns, if present, the desired value. Finally it
    updates an already existing dict in the list or add a new dict to it

    :param d_included: a dict, as returned by res.json().get("included", {})
    :type d_raw: dict
    :param l_posts: a list with dicts
    :type l_posts: list
    :param post_key: the post field name to extract. Example: 'author_name'
    :type post_key: str
    :param post_value: the post value correspoding to post_key
    :type post_value: str

    :return: post list
    :rtype: list
    """
    elements_current_index = len(l_posts) - 1

    if elements_current_index == -1:
        l_posts.append({post_key: post_value})
    else:
        if post_key not in l_posts[elements_current_index]:
            l_posts[elements_current_index][post_key] = post_value
        else:
            l_posts.append({post_key: post_value})
    return l_posts


def parse_list_raw_urns(l_raw_urns: List[str]) -> List[str]:
    """Iterates a list containing posts URNS and retrieves list of URNs

    :param l_raw_urns: List containing posts URNs
    :type l_raw_posts: list

    :return: List of URNs
    :rtype: list
    """
    l_urns = []
    for i in l_raw_urns:
        l_urns.append(get_urn_from_raw_update(i))
    return l_urns


def parse_list_raw_posts(l_raw_posts: List[Dict], linkedin_base_url: str) -> List[Dict]:
    """Iterates a unsorted list containing post fields and assemble a
    list of dicts, each one of them contains a post

    :param l_raw_posts: Unsorted list containing posts information
    :type l_raw_posts: list
    :param linkedin_base_url: Linkedin URL
    :type linkedin_base_url: str

    :return: List of dicts, each one of them is a post
    :rtype: list
    """
    l_posts: list[dict] = []
    for i in l_raw_posts:
        author_name = get_update_author_name(i)
        if author_name:
            l_posts = append_update_post_field_to_posts_list(
                i, l_posts, "author_name", author_name
            )

        author_profile = get_update_author_profile(i, linkedin_base_url)
        if author_profile:
            l_posts = append_update_post_field_to_posts_list(
                i, l_posts, "author_profile", author_profile
            )

        old = get_update_old(i)
        if old:
            l_posts = append_update_post_field_to_posts_list(i, l_posts, "old", old)

        content = get_update_content(i, linkedin_base_url)
        if content:
            l_posts = append_update_post_field_to_posts_list(
                i, l_posts, "content", content
            )

        url = get_update_url(i, linkedin_base_url)
        if url:
            l_posts = append_update_post_field_to_posts_list(i, l_posts, "url", url)

    return l_posts


def get_list_posts_sorted_without_promoted(
    l_urns: List[str], l_posts: List[Dict]
) -> List[Dict]:
    """Iterates l_urns and looks for corresponding dicts in l_posts matching 'url' key.
    If found, removes this dict from l_posts and appends it to the returned list of posts

    :param l_urns: List of posts URNs
    :type l_urns: list
    :param l_posts: List of dicts, which each of them is a post
    :type l_posts: list

    :return: List of dicts, each one of them is a post
    :rtype: list
    """
    l_posts_sorted_without_promoted = []
    l_posts[:] = [d for d in l_posts if "Promoted" not in d.get("old", "")]
    for urn in l_urns:
        for post in l_posts:
            if urn in post["url"]:
                l_posts_sorted_without_promoted.append(post)
                l_posts[:] = [d for d in l_posts if urn not in d.get("url", "")]
                break
    return l_posts_sorted_without_promoted


def generate_trackingId_as_charString() -> str:
    """Generates and returns a random trackingId

    :return: Random trackingId string
    :rtype: str
    """
    random_int_array = [random.randrange(256) for _ in range(16)]
    rand_byte_array = bytearray(random_int_array)
    return "".join([chr(i) for i in rand_byte_array])


def generate_trackingId() -> str:
    """Generates and returns a random trackingId

    :return: Random trackingId string
    :rtype: str
    """
    random_int_array = [random.randrange(256) for _ in range(16)]
    rand_byte_array = bytearray(random_int_array)
    return str(base64.b64encode(rand_byte_array))[2:-1]



"""
MIT License

Copyright (c) 2024 Tom Quirk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from enum import IntEnum, Enum


class Experience(IntEnum):
    INTERN = 1
    ENTRY = 2
    ASSOCIATE = 3
    MID_SENIOR = 4
    DIRECTOR = 5
    EXECUTIVE = 6


class NetworkDepth(Enum):
    FIRST = "F"
    SECOND = "S"
    OTHER = "O"


class JobType(Enum):
    FULL_TIME = "F"
    CONTRACT = "C"
    PART_TIME = "P"
    TEMPORARY = "T"
    INTERN = "I"
    VOLUNTEER = "V"
    OTHER = "O"


class LocationType(IntEnum):
    ONSITE = 1
    REMOTE = 2
    HYBRID = 3


class SortOrder(Enum):
    RELEVANCE = "RELEVANCE"
    LATEST = "REVERSE_CHRONOLOGICAL"


class SortBy(Enum):
    RELEVANCE = "R"
    DATE = "DD"


class CompanyID(IntEnum):
    GOOGLE = 1441
    APPLE = 162479


class JobTitle(IntEnum):
    SOFTWARE_ENGINEER = 9
    CLOUD_ENGINEER = 3006


class GeoID(IntEnum):
    # Countries
    USA = 103644278
    CANADA = 101174742

    # Cities
    ATLANTA = 106224388
    AUSTIN = 104472865
    BALTIMORE = 106330734
    BOSTON = 102380872
    CHARLOTTE = 102264677
    CHICAGO = 103112676
    COLUMBUS = 102812094
    DALLAS = 104194190
    DC = 104383890
    DENVER = 103736294
    DETROIT = 103624908
    FORT_WORTH = 100432370
    HOUSTON = 103743442
    KANSAS_CITY_KANSAS = 104388316
    KANSAS_CITY_MISSOURI = 106142749
    LOS_ANGELES = 102448103
    MIAMI = 102394087
    MINNEAPOLIS = 103039849
    MONTREAL = 101330853
    NEW_YORK = 105080838
    OTTAWA = 106234700
    PHILLY = 104937023
    PHOENIX = 100219842
    PORTLAND = 104727230
    RALEIGH = 100197101
    SALT_LAKE_CITY = 103250458
    SAN_DIEGO = 103918656
    SAN_FRANCISCO = 102277331
    SEATTLE = 104116203
    ST_LOUIS = 104428936
    ST_PAUL = 102370339
    TAMPA = 105517665
    TORONTO = 100025096
    VANCOUVER = 103366113

    # Metros
    ATLANTA_METRO = 90000052
    AUSTIN_METRO = 90000064
    BOSTON_METRO = 90000007
    CHARLOTTE_METRO = 90000152
    CHICAGO_METRO = 90000014
    COLUMBUS_METRO = 90000184
    DALLAS_METRO = 90000031
    DC_METRO = 90000097
    DENVER_METRO = 90000034
    DETROIT_METRO = 90000035
    HOUSTON_METRO = 90000042
    KANSAS_CITY_METRO = 90000376
    LOS_ANGELES_METRO = 90000049
    MIAMI_METRO = 90000056
    MINNEAPOLIS_METRO = 90000512
    MONTREAL_METRO = 90009540
    NEW_YORK_METRO = 90000070
    OTTAWA_METRO = 90009542
    PHILLY_METRO = 90000077
    PHOENIX_METRO = 90000620
    PORTLAND_METRO = 90000079
    RALEIGH_METRO = 90000664
    SALT_LAKE_CITY_METRO = 90000716
    SAN_DIEGO_METRO = 90010472
    SAN_FRANCISCO_METRO = 90000084
    SEATTLE_METRO = 90000091
    ST_LOUIS_METRO = 90000704
    TAMPA_METRO = 90000828
    TORONTO_METRO = 90009551
    VANCOUVER_METRO = 90009553



"""
Provides linkedin api-related code
"""

import http
from typing import Optional
from httpx import Cookies

from li_scrapi.utils import query_options
from li_scrapi.utils._base_linkedin import BaseLinkedIn
from li_scrapi.client import AsyncLinkedInClient
from li_scrapi.utils.schemas import (
    LinkedInProfile,
    LinkedInPrivacySettings,
    LinkedInMemberBadges,
    LinkedInNetwork,
    LinkedInOrganizationResponse,
    LinkedInContactInfo,
    LinkedInSearchPeopleResponse,
    LinkedInProfilePostsResponse,
    LinkedInProfileSkillsResponse,
    LinkedInPostCommentResponse,
    LinkedInSearchCompaniesResponse,
    LinkedInUpdatesResponse,
    LinkedInSelfProfile,
    LinkedInJobSearchResponse,
    LinkedInJob,
    LinkedInJobSkills,
)


class AsyncLinkedIn(BaseLinkedIn):
    """
    Class for accessing the LinkedIn API.

    :param username: Username of LinkedIn account.
    :type username: str
    :param password: Password of LinkedIn account.
    :type password: str
    """

    def __init__(
        self,
        client: AsyncLinkedInClient,
    ):
        """Constructor method"""
        self.client = client
        self._metadata: dict[str, LinkedInSelfProfile] = {}

    async def _close(self):
        await self.client.close()

    async def _fetch(self, uri: str, base_request=False, **kwargs):
        """GET request to Linkedin API"""

        url = f"{self.client.API_BASE_URL if not base_request else self.client.LINKEDIN_BASE_URL}{uri}"
        return await self.client.get(url, **kwargs)

    async def _post(self, uri: str, base_request=False, **kwargs):
        """POST request to Linkedin API"""

        url = f"{self.client.API_BASE_URL if not base_request else self.client.LINKEDIN_BASE_URL}{uri}"
        return await self.client.post(url, **kwargs)

    async def authenticate(
        self, username: str, password: str, cookies: Optional[Cookies] = None
    ):
        if not cookies or not self.client._cookie_repository._is_token_still_valid(
            cookies.jar
        ):
            await self.client.authenticate(username, password)
        else:
            self.client._set_session_cookies(cookies)
            self.client._cookie_repository.save(cookies, username)

    async def get_profile_posts(
        self, urn_id: str, post_count=10, start=0, pagination_token=""
    ) -> LinkedInProfilePostsResponse | None:
        url, url_params = self._api_builder.build_profile_posts_url_params(
            urn_id,
            count=min(post_count, self._MAX_POST_COUNT),
            start=start,
            pagination_token=pagination_token,
        )
        linkedin_response = await self._fetch(url, params=url_params)
        data = linkedin_response.json()
        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get profile posts data, response: {linkedin_response.status_code}"
            )
            return None
        return LinkedInProfilePostsResponse(**data)

    async def get_post_comments(
        self,
        social_detail_urn: str,
        sort_by: query_options.SortOrder = query_options.SortOrder.RELEVANCE,
        comment_count=100,
        start=0,
        pagination_token="",
    ) -> LinkedInPostCommentResponse | None:
        url, url_params = self._api_builder.build_post_comments_url_query(
            post_urn=social_detail_urn,
            count=min(comment_count, self._MAX_POST_COUNT),
            start=start,
            sort_order=sort_by.value,
            pagination_token=pagination_token,
        )
        linkedin_response = await self._fetch(url, params=url_params)
        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get post comments data, response: {linkedin_response.status_code}"
            )
            return None
        data = linkedin_response.json()

        return LinkedInPostCommentResponse(**data)

    async def search(self, params: dict, offset=0) -> dict:
        url, _ = self._api_builder.build_search_url_params_query(params, offset)

        linkedin_response = await self._fetch(url)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get search data, response: {linkedin_response.status_code}"
            )
            return {}

        data = linkedin_response.json()

        paging = self._parse_search_data(data)

        self.logger.debug(f"results grew to {len(paging.get("elements", []))}")

        return paging

    async def search_people(
        self,
        keywords: str = "",
        connection_of: str = "",
        network_depths: list[query_options.NetworkDepth] = [],
        current_company: list[query_options.CompanyID] = [],
        past_companies: list[query_options.CompanyID] = [],
        nonprofit_interests: list[str] = [],
        profile_languages: list[str] = [],
        regions: list[query_options.GeoID] = [],
        industries: list[str] = [],
        schools: list[str] = [],
        service_categories: list[str] = [],
        include_private_profiles=False,  # profiles without a public id, "Linkedin Member"
        keyword_first_name: str = "",
        keyword_last_name: str = "",
        keyword_title: str = "",
        keyword_company: str = "",
        keyword_school: str = "",
        offset: int = 0,
    ) -> LinkedInSearchPeopleResponse:
        kwargs = {
            "connection_of": connection_of,
            "network_depths": [depth.value for depth in network_depths],
            "regions": [f"{region.value}" for region in regions],
            "industries": industries,
            "current_company": [f"{company.value}" for company in current_company],
            "past_companies": [f"{company.value}" for company in past_companies],
            "profile_languages": profile_languages,
            "nonprofit_interests": nonprofit_interests,
            "schools": schools,
            "service_categories": service_categories,
            "keyword_first_name": keyword_first_name,
            "keyword_last_name": keyword_last_name,
            "keyword_title": keyword_title,
            "keyword_company": keyword_company,
            "keyword_school": keyword_school,
        }

        params = self._api_builder.build_search_people_params(**kwargs)

        if keywords:
            params["keywords"] = keywords

        data = await self.search(params, offset=offset)

        results = self._normalize_search_people_data(data, include_private_profiles)
        data["elements"] = results
        return LinkedInSearchPeopleResponse(**data)

    async def search_companies(
        self, keywords: str = "", offset: int = 0
    ) -> LinkedInSearchCompaniesResponse:
        params = self._api_builder.build_search_companies_params(keywords)

        data = await self.search(params)

        results = self._normalize_search_company_data(data)
        data["elements"] = results

        return LinkedInSearchCompaniesResponse(**data)

    async def search_jobs(
        self,
        keywords: str = "",
        companies: list[query_options.CompanyID] = [],
        experience: list[query_options.Experience] = [],
        job_type: list[query_options.JobType] = [],
        job_title: list[query_options.JobTitle] = [],
        industries: list[str] = [],
        location: query_options.GeoID | None = None,
        remote: list[query_options.LocationType] = [],
        listed_at=24 * 60 * 60,
        distance: int | None = None,
        sort_by: query_options.SortBy = query_options.SortBy.RELEVANCE,
        v2: bool = False,
        limit=10,
        offset=0,
    ) -> LinkedInJobSearchResponse | None:
        count = min(limit, self._MAX_SEARCH_COUNT)

        kwargs = {
            "keywords": keywords,
            "experience": [f"{exp.value}" for exp in experience],
            "jobType": [job.value for job in job_type],
            "workplaceType": [f"{rem.value}" for rem in remote],
            "locationUnion": f"{location.value}" if location else "",
            "company": [f"{company.value}" for company in companies],
            "title": [f"{title.value}" for title in job_title],
            "industry": industries,
            "distance": distance,
            "sortBy": sort_by.value,
            "timePostedRange": listed_at,
        }

        url, _ = self._api_builder.build_search_jobs_url_query(count, offset, **kwargs)

        headers = {}
        if v2:
            headers = self._api_builder.get_v2_headers()

        linkedin_response = await self._fetch(url, headers=headers, timeout=10.0)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get jobs, response: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()

        paging = self._normalize_search_jobs_data(data, v2)

        return LinkedInJobSearchResponse(**paging)

    async def get_profile_contact_info(
        self, public_id: str = "", urn_id: str = ""
    ) -> LinkedInContactInfo | None:
        if not public_id and not urn_id:
            self.logger.info("Need to specifiy either a public id or a urn id")
            return None

        url = self._api_builder.build_contact_info_url(public_id, urn_id)
        res = await self._fetch(url)
        data = res.json()

        contact_info = self._normalize_contact_info(data)

        return LinkedInContactInfo(**contact_info)

    async def get_profile_skills(
        self, public_id: str = "", urn_id: str = ""
    ) -> LinkedInProfileSkillsResponse | None:
        if not public_id and not urn_id:
            self.logger.info("Need either public id or urn id to continue")
            return None

        params = self._api_builder.get_default_params()
        url = self._api_builder.build_skills_info_url(public_id, urn_id)
        linkedin_response = await self._fetch(url, params=params)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get profile skills, response: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()

        return LinkedInProfileSkillsResponse(**data)

    async def get_profile(
        self, public_id: str = "", urn_id: str = ""
    ) -> LinkedInProfile | None:
        if not public_id and not urn_id:
            self.logger.info("Need either public id or urn id to continue")
            return None

        url = self._api_builder.build_get_profile_url(public_id, urn_id)
        linkedin_response = await self._fetch(url)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get profile data, response: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()

        # massage [profile] data
        profile = self._normalize_person_data(data)

        return LinkedInProfile(**profile)

    async def get_profile_connections(
        self, urn_id: str, offset: int = 0
    ) -> LinkedInSearchPeopleResponse:
        return await self.search_people(
            connection_of=urn_id,
            network_depths=[query_options.NetworkDepth.FIRST],
            offset=offset,
        )

    async def get_company_updates(
        self,
        public_id: str = "",
        urn_id: str = "",
        start: int = 0,
        count: int = BaseLinkedIn._MAX_UPDATE_COUNT,
    ) -> LinkedInUpdatesResponse | None:
        if not public_id and not urn_id:
            self.logger.info("Need either public id or urn id to continue")
            return None

        url, params = self._api_builder.build_company_updates_url_params(
            public_id, urn_id, min(count, self._MAX_UPDATE_COUNT), start
        )

        linkedin_response = await self._fetch(url, params=params)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"Failed to collect company data, error code: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()
        paging = data["paging"]
        paging["elements"] = data["elements"]

        self.logger.debug(f"results grew: {len(data["elements"])}")

        return LinkedInUpdatesResponse(paging=data["paging"], elements=data["elements"])

    async def get_profile_updates(
        self,
        public_id: str = "",
        urn_id: str = "",
        start: int = 0,
        count: int = BaseLinkedIn._MAX_UPDATE_COUNT,
    ):
        if not public_id and not urn_id:
            self.logger.info("Need either public id or urn id to continue")
            return None

        url, params = self._api_builder.build_profile_updates_url_params(
            public_id, urn_id, min(count, self._MAX_UPDATE_COUNT), start
        )

        linkedin_response = await self._fetch(url, params=params)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get profile updates, response: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()
        return LinkedInUpdatesResponse(**data)

    async def get_organization(self, public_id: str):
        url, params = self._api_builder.build_get_organization_url_params(public_id)

        linkedin_response = await self._fetch(url, params=params)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get organization data, response: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()
        paging = data["paging"]
        elements = data["elements"][0]

        return LinkedInOrganizationResponse(paging=paging, elements=elements)

    async def get_user_profile(self, use_cache=True) -> LinkedInSelfProfile | None:
        me_profile = self._metadata.get("me")
        if not me_profile and (not self.client.metadata.get("me") or not use_cache):
            linkedin_response = await self._fetch("/me")

            if linkedin_response.status_code != http.HTTPStatus.OK:
                self.logger.warning(
                    f"failed to get user profile data, response: {linkedin_response.status_code}"
                )
                return me_profile

            me_profile = LinkedInSelfProfile(**linkedin_response.json())
            # cache profile
            self._metadata["me"] = me_profile

        return me_profile

    async def get_profile_privacy_settings(self, public_profile_id: str):
        res = await self._fetch(
            self._api_builder.build_privacy_settings_url(public_profile_id),
            headers=self._api_builder.get_v2_headers(),
        )
        if res.status_code != http.HTTPStatus.OK:
            return None

        data = res.json()
        return LinkedInPrivacySettings(**data.get("data", {}))

    async def get_profile_member_badges(self, public_profile_id: str):
        res = await self._fetch(
            self._api_builder.build_member_badges_url(public_profile_id),
            headers=self._api_builder.get_v2_headers(),
        )
        if res.status_code != http.HTTPStatus.OK:
            return None

        data = res.json()
        return LinkedInMemberBadges(**data.get("data", {}))

    async def get_profile_network_info(self, public_profile_id: str):
        res = await self._fetch(
            self._api_builder.build_profile_network_url(public_profile_id),
            headers=self._api_builder.get_v2_headers(),
        )
        if res.status_code != http.HTTPStatus.OK:
            return None

        data = res.json()
        return LinkedInNetwork(**data.get("data", {}))

    async def get_job(self, job_id: str) -> LinkedInJob | None:
        url, params = self._api_builder.build_get_job_url_params(job_id)

        linkedin_response = await self._fetch(url, params=params)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get job data, response: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()
        data["workplaceTypes"] = [
            query_options.LocationType(int(val.split(":")[-1]))
            for val in data["workplaceTypes"]
        ]

        return LinkedInJob(**data)

    async def get_job_skills(self, job_id: str) -> LinkedInJobSkills | None:
        url, params = self._api_builder.build_get_job_skill_url_params(job_id)
        linkedin_response = await self._fetch(
            url,
            params=params,
        )
        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get job data, response: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()

        return LinkedInJobSkills(**data)



from pydantic import BaseModel, ConfigDict, Field, HttpUrl, EmailStr
from pydantic_extra_types.country import CountryShortName, CountryAlpha2
from typing import Any
from enum import Enum
from datetime import datetime
from li_scrapi.utils.query_options import LocationType


class ContentSource(Enum):
    JOBS_PREMIUM_OFFLINE = "JOBS_PREMIUM_OFFLINE"
    JOBS_PREMIUM = "JOBS_PREMIUM"
    JOBS_CREATE = "JOBS_CREATE"


class JobState(Enum):
    LISTED = "LISTED"


class ContentType(Enum):
    UNIVERSAL = "UNIVERSAL"
    CONTENT_B = "CONTENT_B"


class DistanceValue(Enum):
    DISTANCE_1 = "DISTANCE_1"
    DISTANCE_2 = "DISTANCE_2"
    DISTANCE_3 = "DISTANCE_3"
    OUT_OF_NETWORK = "OUT_OF_NETWORK"


class EnabledIndicators(Enum):
    ALL_ENABLED = "ALL_ENABLED"
    # TODO add more later as we discover them


class VisibilitySettings(Enum):
    PUBLIC = "PUBLIC"
    # TODO add more later as we discover them


class Base(BaseModel):
    model_config = ConfigDict(extra="ignore")


class Metadata(Base):
    pagination_token: str | None = Field(alias="paginationToken", default=None)


class MiniProfile(Base):
    # TODO see if we can add this to all the other requests that send back a mini profile
    first_name: str = Field(alias="firstName")
    last_name: str = Field(alias="lastName")
    occupation: str | None = None
    entity_urn: str = Field(alias="entityUrn")
    public_identifier: str = Field(alias="publicIdentifier")


class Paging(Base):
    count: int
    start: int
    total: int | None = None


class Distance(Base):
    value: DistanceValue


class Skill(Base):
    name: str


class Date(Base):
    month: int | None = None
    day: int | None = None
    year: int | None = None


class TimePeriod(Base):
    start_date: Date = Field(alias="startDate")
    end_date: Date | None = Field(alias="endDate", default=None)


class Range(Base):
    start: int
    end: int


class Response(Base):
    paging: Paging


class Company(Base):
    employee_count_range: Range | None = None
    industries: list[str] | None = None


class LinkedInExperience(Base):
    entity_urn: str = Field(alias="entityUrn")
    company_name: str = Field(alias="companyName")
    time_period: TimePeriod | None = Field(alias="timePeriod", default=None)
    description: str | None = None
    company: Company | None = None
    title: str
    company_urn: str | None = Field(alias="companyUrn", default=None)
    company_logo_url: HttpUrl | None = Field(alias="companyLogoUrl", default=None)


class LinkedInEducation(Base):
    entity_urn: str = Field(alias="entityUrn")
    description: str | None = None
    degree_name: str | None = Field(alias="degreeName", default=None)
    school_name: str = Field(alias="schoolName")
    school_urn: str | None = Field(alias="schoolUrn", default=None)


class LinkedInPublication(Base):
    name: str
    description: str
    url: HttpUrl


class LinkedInCertification(Base):
    authority: str
    name: str
    time_period: TimePeriod | None = Field(alias="timePeriod", default=None)
    company: Company | None = None
    display_source: str | None = Field(alias="displaySource", default=None)
    url: HttpUrl | None = None


class LinkedInVolunteer(Base):
    role: str
    company_name: str = Field(alias="companyName")
    time_period: TimePeriod = Field(alias="timePeriod")
    cause: str | None = None
    description: str
    company_urn: str | None = Field(alias="companyUrn", default=None)


class LinkedInHonor(Base): ...


class LinkedInProject(Base):
    time_period: TimePeriod = Field(alias="timePeriod")
    description: str
    title: str


class LinkedInProfile(Base):
    summary: str | None = None
    industry_name: str = Field(alias="industryName")
    last_name: str = Field(alias="lastName")
    location_name: str = Field(alias="locationName")
    student: bool
    birth_date: Date | None = Field(alias="birthDate", default=None)
    country: CountryShortName = Field(alias="geoCountryName")
    country_urn: str = Field(alias="geoCountryUrn")
    industry_urn: str = Field(alias="industryUrn")
    first_name: str = Field(alias="firstName")
    entity_urn: str = Field(alias="entityUrn")
    geo_location_name: str = Field(alias="geoLocationName")
    headline: str
    display_picture_url: HttpUrl | None = Field(alias="displayPictureUrl", default=None)
    urn_id: str
    profile_urn: str
    member_urn: str
    public_id: str
    experience: list[LinkedInExperience]
    education: list[LinkedInEducation]
    languages: list[Skill]
    publications: list[LinkedInPublication]
    certifications: list[LinkedInCertification]
    volunteer: list[LinkedInVolunteer]
    honors: list[Any]  # TODO Replace this with actual fields
    projects: list[LinkedInProject]
    skills: list[Skill]


class LinkedInPrivacySettings(Base):
    messaging_type_indicators: EnabledIndicators = Field(
        alias="messagingTypingIndicators"
    )
    allow_open_profile: bool = Field(alias="allowOpenProfile")
    profile_picture_visibility_setting: VisibilitySettings = Field(
        alias="profilePictureVisibilitySetting"
    )
    entity_urn: str = Field(alias="entityUrn")
    show_public_profile: bool = Field(alias="showPublicProfile")
    show_premium_subscriber_badge: bool = Field(alias="showPremiumSubscriberBadge")
    public_profile_picture_visibility_setting: VisibilitySettings = Field(
        alias="publicProfilePictureVisibilitySetting"
    )
    former_name_visibility_setting: VisibilitySettings = Field(
        alias="formerNameVisibilitySetting"
    )
    messaging_seen_receipts: EnabledIndicators = Field(alias="messagingSeenReceipts")
    allow_profile_edit_broadcasts: bool = Field(alias="allowProfileEditBroadcasts")


class LinkedInMemberBadges(Base):
    premium: bool
    influencer: bool
    entity_urn: str = Field(alias="entityUrn")
    open_link: bool = Field(alias="openLink")
    job_seeker: bool = Field(alias="jobSeeker")


class LinkedInNetwork(Base):
    distance: Distance
    entity_urn: str = Field(alias="entityUrn")
    following: bool
    followable: bool
    following_info: str = Field(alias="*followingInfo")
    followers_count: int = Field(alias="followersCount")
    connections_count: int = Field(alias="connectionsCount")


class CompanyIndustry(Base):
    localized_name: str = Field(alias="localizedName")


class OfficeLocation(Base):
    country: CountryAlpha2
    geographic_area: str = Field(alias="geographicArea")
    city: str
    postal_code: str | None = Field(alias="postalCode", default=None)
    line_1: str = Field(alias="line1")


class LinkedInOrganization(Base):
    staffing_company: bool = Field(alias="staffingCompany")
    staff_count: int = Field(alias="staffCount")
    ads_rule: str = Field(alias="adsRule")
    claimable: bool
    lcp_treatment: bool = Field(alias="lcpTreatment")
    name: str
    description: str
    paid_company: bool = Field(alias="paidCompany")
    company_page_url: HttpUrl = Field(alias="companyPageUrl")
    url: HttpUrl
    job_search_page_url: HttpUrl = Field(alias="jobSearchPageUrl")
    specialities: list[str]
    company_industries: list[CompanyIndustry] = Field(alias="companyIndustries")
    headquarter: OfficeLocation
    confirmed_locations: list[OfficeLocation] = Field(alias="confirmedLocations")


class LinkedInOrganizationResponse(Response):
    elements: LinkedInOrganization


class TwitterContactInfo(Base):
    name: str
    credential_id: str = Field(alias="credentialId")


class WebsiteContactInfo(Base):
    url: HttpUrl
    label: str | None = None


class LinkedInContactInfo(Base):
    email_address: EmailStr | None = None
    birth_date: Date | None = None
    twitter: list[TwitterContactInfo]
    websites: list[WebsiteContactInfo]
    phone_numbers: list[Any]  # TODO figure this out
    ims: Any | None = None  # TODO figure this out


class LinkedInSearchElement(Base):
    template: ContentType


class LinkedInSearchPeopleElement(Base):
    urn_id: str
    distance: DistanceValue
    job_title: str
    location: str
    name: str


class LinkedInSearchCompaniesElement(Base):
    urn_id: str
    name: str
    headline: str
    subline: str


class LinkedInSearchPeopleResponse(Response):
    elements: list[LinkedInSearchPeopleElement]


class LinkedInSearchCompaniesResponse(Response):
    elements: list[LinkedInSearchCompaniesElement]


class LinkedInLikes(Base):
    paging: Paging


class LinkedInSocialDetail(Base):
    urn: str
    total_shares: int = Field(alias="totalShares")
    thread_id: str = Field(alias="threadId")
    likes: LinkedInLikes


class LinkedInText(Base):
    text: str


class LinkedInCommentary(Base):
    num_lines: int = Field(alias="numLines")
    text: LinkedInText


class LinkedInProfilePostElement(Base):
    entity_urn: str = Field(alias="entityUrn")
    social_detail: LinkedInSocialDetail = Field(alias="socialDetail")
    commentary: LinkedInCommentary | None = None


class LinkedInProfilePostsResponse(Response):
    metadata: Metadata
    elements: list[LinkedInProfilePostElement]


class LinkedInProfileSkillsResponse(Response):
    elements: list[Skill]


class LinkedInCommentElement(Base):
    commenter_profile_urn: str = Field(alias="commenterProfileId")
    original_language: str | None = Field(alias="originalLanguage", default=None)
    comment: LinkedInText = Field(alias="commentV2")
    created_time: datetime = Field(alias="createdTime")


class LinkedInPostCommentResponse(Response):
    metadata: Metadata
    elements: list[LinkedInCommentElement]


class LinkedInVoyagerValueV2(Base):
    commentary: LinkedInCommentary | None = None
    social_detail: LinkedInSocialDetail = Field(alias="socialDetail")


class LinkedInUpdateValue(Base):
    value_v2: LinkedInVoyagerValueV2 = Field(
        alias="com.linkedin.voyager.feed.render.UpdateV2"
    )


class LinkedInUpdateElement(Base):
    value: LinkedInUpdateValue


class LinkedInUpdatesResponse(Response):
    elements: list[LinkedInUpdateElement]


class LinkedInSelfProfile(Base):
    plan_id: int = Field(alias="plainId")
    public_contact_info: Any = Field(
        alias="publicContactInfo"
    )  # TODO Figure this out later
    premium_subscriber: bool = Field(alias="premiumSubscriber")
    mini_profile: MiniProfile = Field(alias="miniProfile")


class LinkedInJobSearchElement(Base):
    content_source: ContentSource = Field(alias="contentSource")
    entity_urn: str = Field(alias="entityUrn")
    tracking_urn: str = Field(alias="trackingUrn")
    reposted_job: bool = Field(alias="repostedJob")
    title: str
    poster_id: str | None = Field(alias="posterId", default=None)


class LinkedInJobSearchResponse(Response):
    elements: list[LinkedInJobSearchElement]


class OffsiteApply(Base):
    applyStartersPreferenceVoid: bool | None = None
    company_apply_url: HttpUrl = Field(alias="companyApplyUrl")
    in_page_offsite_apply: bool | None = Field(alias="inPageOffsiteApply", default=None)


class ComplexApply(Base):
    unify_apply_enabled: bool | None = Field(alias="unifyApplyEnabled", default=None)
    company_apply_url: HttpUrl | None = Field(alias="companyApplyUrl", default=None)
    easy_apply_url: HttpUrl = Field(alias="easyApplyUrl")


class LinkedInApplyMethod(Base):
    off_site_apply: OffsiteApply | None = Field(
        default=None, alias="com.linkedin.voyager.jobs.OffsiteApply"
    )
    complex_site_apply: ComplexApply | None = Field(
        default=None, alias="com.linkedin.voyager.jobs.ComplexOnsiteApply"
    )


class LinkedInJob(Base):
    job_id: int = Field(alias="jobPostingId")
    listed_at: datetime = Field(alias="listedAt")
    title: str
    job_state: JobState = Field(alias="jobState")
    description: LinkedInText
    work_remote_allowed: bool = Field(alias="workRemoteAllowed")
    formatted_location: str = Field(alias="formattedLocation")
    work_place_types: list[LocationType] = Field(alias="workplaceTypes")
    apply_method: LinkedInApplyMethod = Field(alias="applyMethod")


class SkillMatchStatus(Base):
    localized_skill_display_name: str = Field(alias="localizedSkillDisplayName")
    skill: Skill


class LinkedInJobSkills(Base):
    company_urn: str | None = Field(alias="companyUrn", default=None)
    entity_urn: str = Field(alias="entityUrn")
    skill_match_status: list[SkillMatchStatus] = Field(alias="skillMatchStatuses")



from li_scrapi import LinkedIn
from httpx import Client, Cookies, Proxy
from li_scrapi.client import LinkedInClient
from li_scrapi.utils.schemas import (
    LinkedInProfilePostsResponse,
    LinkedInPostCommentResponse,
    LinkedInUpdatesResponse,
    LinkedInJobSearchResponse,
    LinkedInSearchCompaniesResponse,
    LinkedInSearchPeopleResponse,
)
from li_scrapi.utils.query_options import (
    SortOrder,
    SortBy,
    GeoID,
    LocationType,
    JobTitle,
    JobType,
    Experience,
    CompanyID,
    NetworkDepth,
)


class LinkedInScriptApi:
    def __init__(
        self,
        username: str,
        password: str,
        cookies: Cookies | None = None,
        proxy: Proxy | None = None,
    ) -> None:
        self._linkedin = LinkedIn(
            client=LinkedInClient(session=Client(cookies=cookies, proxy=proxy))
        )
        self._linkedin.authenticate(username, password)

    def get_profile(self, public_id="", urn_id=""):
        return self._linkedin.get_profile(public_id=public_id, urn_id=urn_id)

    def get_posts_from_profile(
        self, urn_id: str, total_posts: int = 10
    ) -> list[LinkedInProfilePostsResponse]:
        res: list[LinkedInProfilePostsResponse] = []
        start = 0
        num_posts = total_posts
        current_posts = self._linkedin.get_profile_posts(urn_id, post_count=total_posts)
        if current_posts and current_posts.elements:
            res.append(current_posts)
        if total_posts > self._linkedin._MAX_POST_COUNT:
            while current_posts and current_posts.metadata.pagination_token:
                start += current_posts.paging.count
                num_posts -= current_posts.paging.count
                if total_posts <= start:
                    break

                current_posts = self._linkedin.get_profile_posts(
                    urn_id,
                    start=start,
                    post_count=num_posts,
                    pagination_token=current_posts.metadata.pagination_token,
                )
                if current_posts and current_posts.elements:
                    res.append(current_posts)

        return res

    def get_post_comments_from_post(
        self, social_detail_urn: str, sort_order=SortOrder.RELEVANCE, total_comments=100
    ) -> list[LinkedInPostCommentResponse]:
        res: list[LinkedInPostCommentResponse] = []
        start = 0
        num_comments = total_comments
        current_comments = self._linkedin.get_post_comments(
            social_detail_urn=social_detail_urn,
            sort_by=sort_order,
            comment_count=total_comments,
        )
        if current_comments and current_comments.elements:
            res.append(current_comments)
        if total_comments > self._linkedin._MAX_POST_COUNT:
            while current_comments and current_comments.metadata.pagination_token:
                start += current_comments.paging.count
                num_comments -= current_comments.paging.count
                if total_comments <= start:
                    break

                current_comments = self._linkedin.get_post_comments(
                    social_detail_urn=social_detail_urn,
                    sort_by=sort_order,
                    comment_count=num_comments,
                    start=start,
                    pagination_token=current_comments.metadata.pagination_token,
                )
                if current_comments and current_comments.elements:
                    res.append(current_comments)

        return res

    def search_jobs(
        self,
        keywords: str = "",
        companies: list[CompanyID] = [],
        experience: list[Experience] = [],
        job_type: list[JobType] = [],
        job_title: list[JobTitle] = [],
        industries: list[str] = [],
        location: GeoID | None = GeoID.USA,
        remote: list[LocationType] = [],
        listed_at=24 * 60 * 60,
        distance: int | None = None,
        sort_by: SortBy = SortBy.RELEVANCE,
        v2: bool = False,
        total_jobs=10,
    ) -> list[LinkedInJobSearchResponse]:
        res: list[LinkedInJobSearchResponse] = []
        kwargs = {
            "keywords": keywords,
            "experience": experience,
            "job_type": job_type,
            "companies": companies,
            "job_title": job_title,
            "industries": industries,
            "location": location,
            "remote": remote,
            "listed_at": listed_at,
            "distance": distance,
            "sort_by": sort_by,
            "v2": v2,
        }
        start = 0
        num_jobs = total_jobs
        all_jobs = self._linkedin.search_jobs(**kwargs, limit=total_jobs)
        if all_jobs and all_jobs.elements:
            res.append(all_jobs)
        if total_jobs > self._linkedin._MAX_SEARCH_COUNT:
            while all_jobs and all_jobs.elements:
                start += all_jobs.paging.count
                num_jobs -= all_jobs.paging.count
                if total_jobs <= start:
                    break

                all_jobs = self._linkedin.search_jobs(
                    **kwargs, limit=num_jobs, offset=start
                )
                if all_jobs and all_jobs.elements:
                    res.append(all_jobs)
        return res

    def search_people(
        self,
        keywords: str = "",
        connection_of: str = "",
        network_depths: list[NetworkDepth] = [],
        current_company: list[CompanyID] = [],
        past_companies: list[CompanyID] = [],
        nonprofit_interests: list[str] = [],
        profile_languages: list[str] = [],
        regions: list[GeoID] = [],
        industries: list[str] = [],
        schools: list[str] = [],
        service_categories: list[str] = [],
        include_private_profiles: bool = False,
        keyword_first_name: str = "",
        keyword_last_name: str = "",
        keyword_title: str = "",
        keyword_company: str = "",
        keyword_school: str = "",
        total_people: int = 10,
    ) -> list[LinkedInSearchPeopleResponse]:
        res: list[LinkedInSearchPeopleResponse] = []
        kwargs = {
            "keywords": keywords,
            "connection_of": connection_of,
            "network_depths": network_depths,
            "current_company": current_company,
            "past_companies": past_companies,
            "nonprofit_interests": nonprofit_interests,
            "profile_languages": profile_languages,
            "regions": regions,
            "industries": industries,
            "schools": schools,
            "service_categories": service_categories,
            "include_private_profiles": include_private_profiles,
            "keyword_first_name": keyword_first_name,
            "keyword_last_name": keyword_last_name,
            "keyword_title": keyword_title,
            "keyword_company": keyword_company,
            "keyword_school": keyword_school,
            "offset": 0,
        }
        start = 0
        num_people = total_people
        all_people = self._linkedin.search_people(**kwargs)
        if all_people.elements:
            res.append(all_people)
        if total_people > all_people.paging.count:
            while all_people and all_people.elements:
                start += all_people.paging.count
                num_people -= all_people.paging.count
                if total_people <= start:
                    break
                kwargs["offset"] = start
                all_people = self._linkedin.search_people(**kwargs)
                if all_people.elements:
                    res.append(all_people)

        return res

    def search_companies(
        self, keywords: str = "", total_companies=10
    ) -> list[LinkedInSearchCompaniesResponse]:
        res: list[LinkedInSearchCompaniesResponse] = []
        all_companies = self._linkedin.search_companies(keywords)
        start = 0
        num_companies = total_companies
        if all_companies.elements:
            res.append(all_companies)
        if total_companies > all_companies.paging.count:
            while all_companies and all_companies.elements:
                start += all_companies.paging.count
                num_companies -= all_companies.paging.count
                if total_companies <= start:
                    break

                all_companies = self._linkedin.search_companies(
                    keywords=keywords, offset=start
                )
                if all_companies.elements:
                    res.append(all_companies)
        return res

    def get_contact_info_from_profile(self, public_id="", urn_id=""):
        return self._linkedin.get_profile_contact_info(
            public_id=public_id, urn_id=urn_id
        )

    def get_skills_from_profile(self, public_id="", urn_id=""):
        return self._linkedin.get_profile_skills(public_id=public_id, urn_id=urn_id)

    def get_connections_from_profile(
        self, urn_id: str = "", total_connections: int = 10
    ) -> list[LinkedInSearchPeopleResponse]:
        res: list[LinkedInSearchPeopleResponse] = []
        start = 0
        num_people = total_connections
        all_connections = self._linkedin.get_profile_connections(urn_id=urn_id)
        if all_connections.elements:
            res.append(all_connections)
        if total_connections > all_connections.paging.count:
            while all_connections and all_connections.elements:
                start += all_connections.paging.count
                num_people -= all_connections.paging.count
                if total_connections <= start:
                    break
                all_connections = self._linkedin.get_profile_connections(
                    urn_id=urn_id, offset=start
                )
                if all_connections.elements:
                    res.append(all_connections)

        return res

    def get_updates_from_company(self, public_id="", urn_id="", total_posts=10):
        res: list[LinkedInUpdatesResponse] = []
        start = 0
        num_posts = total_posts
        all_updates = self._linkedin.get_company_updates(
            public_id=public_id, urn_id=urn_id, count=total_posts
        )
        if all_updates and all_updates.elements:
            res.append(all_updates)
        if total_posts > self._linkedin._MAX_UPDATE_COUNT:
            while all_updates and all_updates.elements:
                start += all_updates.paging.count
                num_posts -= all_updates.paging.count
                if total_posts <= start:
                    break

                all_updates = self._linkedin.get_company_updates(
                    public_id=public_id, urn_id=urn_id, start=start, count=num_posts
                )
                if all_updates and all_updates.elements:
                    res.append(all_updates)

        return res

    def get_updates_from_profile(self, public_id="", urn_id="", total_posts=10):
        res: list[LinkedInUpdatesResponse] = []
        start = 0
        num_posts = total_posts
        all_updates = self._linkedin.get_profile_updates(
            public_id=public_id, urn_id=urn_id, count=total_posts
        )
        if all_updates and all_updates.elements:
            res.append(all_updates)
        if total_posts > self._linkedin._MAX_UPDATE_COUNT:
            while all_updates and all_updates.elements:
                start += all_updates.paging.count
                num_posts -= all_updates.paging.count
                if total_posts <= start:
                    break

                all_updates = self._linkedin.get_profile_updates(
                    public_id=public_id, urn_id=urn_id, start=start, count=num_posts
                )
                if all_updates and all_updates.elements:
                    res.append(all_updates)

        return res

    def get_organization(self, public_id: str):
        return self._linkedin.get_organization(public_id)

    def get_user_profile(self, use_cache=True):
        return self._linkedin.get_user_profile(use_cache=use_cache)

    def get_privacy_settings_from_profile(self, public_profile_id: str):
        return self._linkedin.get_profile_privacy_settings(public_profile_id)

    def get_network_from_profile(self, public_profile_id: str):
        return self._linkedin.get_profile_network_info(public_profile_id)

    def get_job(self, job_id: str):
        return self._linkedin.get_job(job_id)

    def get_job_skills(self, job_id: str):
        return self._linkedin.get_job_skills(job_id)



"""
MIT License

Copyright (c) 2024 Tom Quirk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from httpx import _client, Cookies, Response
from bs4 import BeautifulSoup, Tag
from li_scrapi.cookie_repository import CookieRepository
from li_scrapi.utils.errors import ChallengeException, UnauthorizedException
from tenacity import (
    retry,
    wait_exponential_jitter,
    stop_after_attempt,
    retry_if_not_result,
)
from abc import ABC, abstractmethod
from aiolimiter import AsyncLimiter

import loguru
import json
import http


class Client(ABC):
    # Settings for general Linkedin API calls
    LINKEDIN_BASE_URL = "https://www.linkedin.com"
    API_BASE_URL = f"{LINKEDIN_BASE_URL}/voyager/api"
    REQUEST_HEADERS = {
        "user-agent": " ".join(
            [
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5)",
                "AppleWebKit/537.36 (KHTML, like Gecko)",
                "Chrome/83.0.4103.116 Safari/537.36",
            ]
        ),
        # "accept": "application/vnd.linkedin.normalized+json+2.1",
        "accept-language": "en-AU,en-GB;q=0.9,en-US;q=0.8,en;q=0.7",
        "x-li-lang": "en_US",
        "x-restli-protocol-version": "2.0.0",
        # "x-li-track": '{"clientVersion":"1.2.6216","osName":"web","timezoneOffset":10,"deviceFormFactor":"DESKTOP","mpName":"voyager-web"}',
    }

    # Settings for authenticating with Linkedin
    AUTH_REQUEST_HEADERS = {
        "X-Li-User-Agent": "LIAuthLibrary:0.0.3 com.linkedin.android:4.1.881 Asus_ASUS_Z01QD:android_9",
        "User-Agent": "ANDROID OS",
        "X-User-Language": "en",
        "X-User-Locale": "en_US",
        "Accept-Language": "en-us",
    }

    logger = loguru.logger
    _cookie_repository = CookieRepository()

    @abstractmethod
    def __init__(
        self, session: _client.BaseClient, cookies_dir: str, refresh_cookies=False
    ):
        self._session: _client.BaseClient
        self.metadata: dict

    @property
    def cookies(self):
        return self._session.cookies

    def _set_session_cookies(self, cookies: Cookies):
        """
        Set cookies of the current session and save them to a file named as the username.
        """
        self._session.cookies = cookies
        if "li_at" in cookies:
            self._session.headers["li_at"] = self._session.cookies["li_at"]
        if "JSESSIONID" in cookies:
            self._session.headers["csrf-token"] = self._session.cookies[
                "JSESSIONID"
            ].strip('"')

    def _parse_metadata(self, soup: BeautifulSoup):
        clientApplicationInstanceRaw = soup.find(
            "meta", attrs={"name": "applicationInstance"}
        )
        if clientApplicationInstanceRaw and isinstance(
            clientApplicationInstanceRaw, Tag
        ):
            raw_content = clientApplicationInstanceRaw.attrs.get("content", {})
            clientApplicationInstance = json.loads(raw_content)
            self.metadata["clientApplicationInstance"] = clientApplicationInstance

        clientPageInstanceIdRaw = soup.find(
            "meta", attrs={"name": "clientPageInstanceId"}
        )
        if clientPageInstanceIdRaw and isinstance(clientPageInstanceIdRaw, Tag):
            clientPageInstanceId = clientPageInstanceIdRaw.attrs.get("content", {})
            self.metadata["clientPageInstanceId"] = clientPageInstanceId

    @staticmethod
    def _has_failed(value):
        Client.logger.debug(
            f"status code of {value.request.url} \n is {value.status_code} if not 200, \n retry ..."
        )
        return value.status_code == http.HTTPStatus.OK

    @abstractmethod
    def get(self, url: str, **kwargs) -> Response: ...

    @abstractmethod
    def post(self, url: str, **kwargs) -> Response: ...

    @abstractmethod
    def authenticate(self, username: str, password: str): ...

    @abstractmethod
    def close(self): ...


class LinkedInClient(Client):
    """
    Class to act as a client for the Linkedin API.
    """

    def __init__(
        self, session: _client.Client, cookies_dir: str = "", refresh_cookies=False
    ):
        self._session: _client.Client = session
        self._session.headers.update(self.REQUEST_HEADERS)
        self.metadata = {}
        self._use_cookie_cache = not refresh_cookies
        self._cookie_repository.set_cookies_dir(cookies_dir)

    def close(self):
        self._session.close()

    def _request_session_cookies(self):
        """
        Return a new set of session cookies as given by Linkedin.
        """
        self.logger.debug("Requesting new cookies.")
        res = self.get(
            f"{self.LINKEDIN_BASE_URL}/uas/authenticate",
            headers=self.AUTH_REQUEST_HEADERS,
        )
        return res.cookies

    def authenticate(self, username: str, password: str):
        if self._use_cookie_cache:
            self.logger.debug("Attempting to use cached cookies")
            cookies = self._cookie_repository.get(username)
            if cookies:
                self.logger.debug("Using cached cookies")
                self._set_session_cookies(cookies)
                self._fetch_metadata()
                return

        self._do_authentication_request(username, password)
        self._fetch_metadata()

    def _fetch_metadata(self):
        """
        Get metadata about the "instance" of the LinkedIn application for the signed in user.

        Store this data in self.metadata
        """
        res = self._session.get(
            f"{self.LINKEDIN_BASE_URL}",
            cookies=self._session.cookies,
            headers=self.AUTH_REQUEST_HEADERS,
        )

        soup = BeautifulSoup(res.text, "lxml")
        self._parse_metadata(soup)

    def _do_authentication_request(self, username: str, password: str):
        """
        Authenticate with Linkedin.

        Return a session object that is authenticated.
        """
        self._set_session_cookies(self._request_session_cookies())

        payload = {
            "session_key": username,
            "session_password": password,
            "JSESSIONID": self._session.cookies["JSESSIONID"],
        }

        res = self.post(
            f"{self.LINKEDIN_BASE_URL}/uas/authenticate",
            data=payload,
            headers=self.AUTH_REQUEST_HEADERS,
            cookies=self._session.cookies,
        )

        if res.status_code > http.HTTPStatus.OK:
            raise UnauthorizedException()

        data = res.json()

        if data and data["login_result"] != "PASS":
            raise ChallengeException(data["login_result"])

        self._set_session_cookies(res.cookies)
        res.request.url
        self._cookie_repository.save(res.cookies, username)

    @retry(
        wait=wait_exponential_jitter(2, 15),
        stop=stop_after_attempt(5),
        retry=retry_if_not_result(Client._has_failed),
        retry_error_callback=lambda x: x.outcome.result(),  # type: ignore
    )  # type: ignore
    def post(self, url: str, **kwargs):
        return self._session.post(url, **kwargs)

    @retry(
        wait=wait_exponential_jitter(2, 15),
        stop=stop_after_attempt(5),
        retry=retry_if_not_result(Client._has_failed),
        retry_error_callback=lambda x: x.outcome.result(),  # type: ignore
    )  # type: ignore
    def get(self, url: str, **kwargs):
        return self._session.get(url, **kwargs)


class AsyncLinkedInClient(Client):
    def __init__(
        self, session: _client.AsyncClient, cookies_dir: str = "", refresh_cookies=False
    ):
        self._session: _client.AsyncClient = session
        self._session.headers.update(self.REQUEST_HEADERS)
        self.metadata = {}
        self._use_cookie_cache = not refresh_cookies
        self._cookie_repository.set_cookies_dir(cookies_dir)
        self._limiter = AsyncLimiter(200, 60)

    async def close(self):
        await self._session.aclose()

    async def _request_session_cookies(self):
        """
        Return a new set of session cookies as given by Linkedin.
        """
        self.logger.debug("Requesting new cookies.")

        res = await self._session.get(
            f"{self.LINKEDIN_BASE_URL}/uas/authenticate",
            headers=self.AUTH_REQUEST_HEADERS,
        )
        return res.cookies

    async def authenticate(self, username: str, password: str):
        if self._use_cookie_cache:
            self.logger.debug("Attempting to use cached cookies")
            cookies = self._cookie_repository.get(username)
            if cookies:
                self.logger.debug("Using cached cookies")
                self._set_session_cookies(cookies)
                await self._fetch_metadata()
                return

        await self._do_authentication_request(username, password)
        await self._fetch_metadata()

    async def _fetch_metadata(self):
        """
        Get metadata about the "instance" of the LinkedIn application for the signed in user.

        Store this data in self.metadata
        """
        res = await self._session.get(
            f"{self.LINKEDIN_BASE_URL}",
            cookies=self._session.cookies,
            headers=self.AUTH_REQUEST_HEADERS,
        )

        soup = BeautifulSoup(res.text, "lxml")
        self._parse_metadata(soup)

    async def _do_authentication_request(self, username: str, password: str):
        """
        Authenticate with Linkedin.

        Return a session object that is authenticated.
        """
        self._set_session_cookies(await self._request_session_cookies())

        payload = {
            "session_key": username,
            "session_password": password,
            "JSESSIONID": self._session.cookies["JSESSIONID"],
        }

        res = await self._session.post(
            f"{self.LINKEDIN_BASE_URL}/uas/authenticate",
            data=payload,
            cookies=self._session.cookies,
            headers=self.AUTH_REQUEST_HEADERS,
        )

        if res.status_code > http.HTTPStatus.OK:
            raise UnauthorizedException()

        data = res.json()

        if data and data["login_result"] != "PASS":
            raise ChallengeException(data["login_result"])

        self._set_session_cookies(res.cookies)
        self._cookie_repository.save(res.cookies, username)

    @retry(
        wait=wait_exponential_jitter(2, 15),
        stop=stop_after_attempt(5),
        retry=retry_if_not_result(Client._has_failed),
        retry_error_callback=lambda x: x.outcome.result(),  # type: ignore
    )  # type: ignore
    async def post(self, url: str, **kwargs):
        async with self._limiter:
            return await self._session.post(url, **kwargs)

    @retry(
        wait=wait_exponential_jitter(2, 15),
        stop=stop_after_attempt(5),
        retry=retry_if_not_result(Client._has_failed),
        retry_error_callback=lambda x: x.outcome.result(),  # type: ignore
    )  # type: ignore
    async def get(self, url: str, **kwargs):
        async with self._limiter:
            return await self._session.get(url, **kwargs)



"""
MIT License

Copyright (c) 2024 Tom Quirk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from urllib.parse import urlencode, quote_plus


class ApiBuilder:
    @staticmethod
    def build_feed_url():
        return "/feed/updates"

    @staticmethod
    def get_default_params(count: int = 100, start: int = 0):
        return {"count": count, "start": start}

    @staticmethod
    def build_get_job_skill_url_params(job_id: str):
        url = f"/voyagerAssessmentsDashJobSkillMatchInsight/{quote_plus(f"urn:li:fsd_jobSkillMatchInsight:{job_id}")}"
        return url, {
            "decorationId": "com.linkedin.voyager.dash.deco.assessments.FullJobSkillMatchInsight-17",
        }

    @staticmethod
    def build_get_job_url_params(job_id: str):
        return f"/jobs/jobPostings/{job_id}", {
            "decorationId": "com.linkedin.voyager.deco.jobs.web.shared.WebLightJobPosting-23",
        }

    @staticmethod
    def get_unfollow_entity_payload(urn_id: str):
        return {"urn": f"urn:li:fs_followingInfo:{urn_id}"}

    @staticmethod
    def build_profile_updates_url_params(
        public_id: str, urn_id: str, count: int, start: int
    ):
        return ApiBuilder.build_feed_url(), {
            "profileId": public_id or urn_id,
            "q": "memberShareFeed",
            "moduleKey": "member-share",
            "count": count,
            "start": start,
        }

    @staticmethod
    def build_company_updates_url_params(
        public_id: str, urn_id: str, count: int, start: int
    ):
        return ApiBuilder.build_feed_url(), {
            "companyUniversalName": public_id or urn_id,
            "q": "companyFeedByUniversalName",
            "moduleKey": "member-share",
            "count": count,
            "start": start,
        }

    @staticmethod
    def build_profile_posts_url_params(
        urn_id: str, count: int, start: int, pagination_token: str
    ):
        params = {
            "count": count,
            "start": start,
            "q": "memberShareFeed",
            "moduleKey": "member-shares:phone",
            "includeLongTermHistory": True,
            "profileUrn": f"urn:li:fsd_profile:{urn_id}",
        }
        if pagination_token:
            params["paginationToken"] = pagination_token
        return "/identity/profileUpdatesV2", params

    @staticmethod
    def build_post_comments_url_query(
        post_urn: str, count: int, start: int, sort_order: str, pagination_token: str
    ):
        params = {
            "count": count,
            "start": start,
            "q": "comments",
            "sortOrder": sort_order,
            "updateId": f"activity:{post_urn}",
        }
        if pagination_token:
            params["paginationToken"] = pagination_token
        return "/feed/comments", params

    @staticmethod
    def build_search_url_params_query(params: dict, start: int):
        default_params = {
            "filters": "List()",
            "origin": "GLOBAL_SEARCH_HEADER",
            "q": "all",
            "start": start,
            "queryContext": "List(spellCorrectionEnabled->true,relatedSearchesEnabled->true,kcardTypes->PROFILE|COMPANY)",
            "includeWebMetadata": "true",
        }
        default_params.update(params)

        keywords = (
            f"keywords:{default_params['keywords']},"
            if "keywords" in default_params
            else ""
        )

        url = (
            f"/graphql?variables=(start:{default_params['start']},origin:{default_params['origin']},"
            f"query:("
            f"{keywords}"
            f"flagshipSearchIntent:SEARCH_SRP,"
            f"queryParameters:{default_params['filters']},"
            f"includeFiltersInResponse:false))&queryId=voyagerSearchDashClusters"
            f".b0928897b71bd00a5a7291755dcd64f0"
        )
        return url, default_params

    @staticmethod
    def build_search_companies_params(keywords: str):
        default_params = {
            "filters": "List((key:resultType,value:List(COMPANIES)))",
            "queryContext": "List(spellCorrectionEnabled->true)",
            "keywords": [keywords],
        }

        if not keywords:
            del default_params["keywords"]

        return default_params

    @staticmethod
    def build_search_people_params(**kwargs):
        filters = ["(key:resultType,value:List(PEOPLE))"]
        ops = {
            "connection_of": "(key:connectionOf,value:List({}))",
            "network_depths": "(key:network,value:List({stringify}))",
            "regions": "(key:geoUrn,value:List({stringify}))",
            "industries": "(key:industry,value:List({stringify}))",
            "current_company": "(key:currentCompany,value:List({stringify}))",
            "past_companies": "(key:pastCompany,value:List({stringify}))",
            "profile_languages": "(key:profileLanguage,value:List({stringify}))",
            "nonprofit_interests": "(key:nonprofitInterest,value:List({stringify}))",
            "schools": "(key:schools,value:List({stringify}))",
            "service_categories": "(key:serviceCategory,value:List({stringify}))",
            "keyword_first_name": "(key:firstName,value:List({}))",
            "keyword_last_name": "(key:lastName,value:List({}))",
            "keyword_title": "(key:title,value:List({}))",
            "keyword_company": "(key:company,value:List({}))",
            "keyword_school": "(key:school,value:List({}))",
        }
        for key, value in kwargs.items():
            if "stringify" in ops[key] and value:
                stringify = " | ".join(value)
                filters.append(ops[key].format(stringify=stringify))
            elif value:
                filters.append(ops[key].format(value))

        return {"filters": f"List({",".join(filters)})"}

    @staticmethod
    def build_search_jobs_url_query(count: int, offset: int, **kwargs):
        """
        Query structure:
         "(
            origin:JOB_SEARCH_PAGE_QUERY_EXPANSION,
            keywords:marketing%20manager,
            locationFallback:germany,
            selectedFilters:(
                distance:List(25),
                company:List(163253),
                salaryBucketV2:List(5),
                timePostedRange:List(r2592000),
                workplaceType:List(1)
            ),
            spellCorrectionEnabled:true
        #  )"
        """
        query = {
            "origin": "JOB_SEARCH_PAGE_QUERY_EXPANSION",
            "selectedFilters": {},
            "spellCorrectionEnabled": "true",
            "keywords": "KEYWORD_PLACEHOLDER",
            "locationUnion": "",
        }

        ops = {
            "experience": "List({stringify})",
            "jobType": "List({stringify})",
            "workplaceType": "List({stringify})",
            "locationUnion": "(geoId:{})",
            "company": "List({stringify})",
            "title": "List({stringify})",
            "industry": "List({stringify})",
            "distance": "List({})",
            "sortBy": "List({})",
            "timePostedRange": "List(r{})",
        }
        for key, value in kwargs.items():
            if key == "keywords":
                continue
            if key == "locationUnion":
                query[key] = ops[key].format(value)
            elif "stringify" in ops[key] and value:
                stringify = ",".join(value)
                query["selectedFilters"][key] = ops[key].format(stringify=stringify)
            elif value:
                query["selectedFilters"][key] = ops[key].format(value)

        query_string = (
            str(query)
            .replace(" ", "")
            .replace("'", "")
            .replace("KEYWORD_PLACEHOLDER", kwargs.get("keywords", ""))
            .replace("{", "(")
            .replace("}", ")")
        )

        default_params = {
            "decorationId": "com.linkedin.voyager.dash.deco.jobs.search.JobSearchCardsCollection-174",
            "count": count,
            "q": "jobSearch",
            "query": query_string,
            "start": offset,
        }

        return (
            f"/voyagerJobsDashJobCards?{urlencode(default_params, safe='(),:')}",
            default_params,
        )

    @staticmethod
    def build_contact_info_url(public_id: str | None, urn_id: str | None):
        return f"/identity/profiles/{public_id or urn_id}/profileContactInfo"

    @staticmethod
    def build_skills_info_url(public_id: str | None, urn_id: str | None):
        return f"/identity/profiles/{public_id or urn_id}/skills"

    @staticmethod
    def build_privacy_settings_url(public_id: str):
        return f"/identity/profiles/{public_id}/privacySettings"

    @staticmethod
    def build_member_badges_url(public_id: str):
        return f"/identity/profiles/{public_id}/memberBadges"

    @staticmethod
    def build_profile_network_url(public_id: str):
        return f"/identity/profiles/{public_id}/networkinfo"

    @staticmethod
    def build_get_profile_url(public_id: str | None, urn_id: str | None):
        # NOTE this still works for now, but will probably eventually have to be converted to
        # https://www.linkedin.com/voyager/api/identity/profiles/ACoAAAKT9JQBsH7LwKaE9Myay9WcX8OVGuDq9Uw
        return f"/identity/profiles/{public_id or urn_id}/profileView"

    @staticmethod
    def build_get_organization_url_params(public_id: str):
        return "/organization/companies", {
            "decorationId": "com.linkedin.voyager.deco.organization.web.WebFullCompanyMain-12",
            "q": "universalName",
            "universalName": public_id,
        }

    @staticmethod
    def get_v2_headers():
        return {"accept": "application/vnd.linkedin.normalized+json+2.1"}



"""
MIT License

Copyright (c) 2024 Tom Quirk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import time
import li_scrapi.utils.settings as settings
from http.cookiejar import CookieJar, MozillaCookieJar
from httpx import Cookies
from li_scrapi.utils.errors import LinkedinSessionExpired


class CookieRepository:
    """
    Class to act as a repository for the cookies.
    """

    def __init__(self):
        self.cookie_jar = MozillaCookieJar()

    def set_cookies_dir(self, cookies_dir=settings.COOKIE_PATH):
        self.cookies_dir = cookies_dir or settings.COOKIE_PATH

    def save(self, cookies: Cookies, username: str):
        self._ensure_cookies_dir()
        cookiejar_filepath = self._get_cookies_filepath(username)
        for cookie in cookies.jar:
            self.cookie_jar.set_cookie(cookie)
        self.cookie_jar.save(cookiejar_filepath)

    def get(self, username: str) -> Cookies:
        cookies = self._load_cookies_from_cache(username)
        if cookies and not CookieRepository._is_token_still_valid(cookies.jar):
            raise LinkedinSessionExpired
        return cookies

    def _ensure_cookies_dir(self):
        if not os.path.exists(self.cookies_dir):
            os.makedirs(self.cookies_dir)

    def _get_cookies_filepath(self, username) -> str:
        """
        Return the absolute path of the cookiejar for a given username
        """
        return "{}{}.txt".format(self.cookies_dir, username)

    def _load_cookies_from_cache(self, username: str) -> Cookies:
        cookie_jar = None
        cookiejar_filepath = self._get_cookies_filepath(username)
        if os.path.exists(cookiejar_filepath):
            self.cookie_jar.load(cookiejar_filepath)
            cookie_jar = self.cookie_jar
        return Cookies(cookie_jar)

    @staticmethod
    def _is_token_still_valid(cookiejar: CookieJar):
        _now = time.time()
        for cookie in cookiejar:
            if cookie.name == "JSESSIONID" and cookie.value:
                if cookie.expires and cookie.expires > _now:
                    return True
            if cookie.name == "li_at" and cookie.value:
                if cookie.expires and cookie.expires > _now:
                    return True

        return False



from .linkedin import LinkedIn
from .async_linkedin import AsyncLinkedIn
from .linkedin_api import LinkedInScriptApi
from .client import AsyncLinkedInClient, LinkedInClient
from .utils.schemas import (
    ContentSource,
    JobState,
    ContentType,
    DistanceValue,
    EnabledIndicators,
    VisibilitySettings,
    MiniProfile,
    LinkedInEducation,
    LinkedInExperience,
    LinkedInPublication,
    LinkedInCertification,
    LinkedInVolunteer,
    LinkedInProject,
    LinkedInProfile,
    LinkedInPrivacySettings,
    LinkedInMemberBadges,
    LinkedInNetwork,
    CompanyIndustry,
    OfficeLocation,
    LinkedInOrganization,
    LinkedInOrganizationResponse,
    TwitterContactInfo,
    WebsiteContactInfo,
    LinkedInContactInfo,
    LinkedInSearchElement,
    LinkedInSearchPeopleElement,
    LinkedInSearchCompaniesElement,
    LinkedInSearchPeopleResponse,
    LinkedInSearchCompaniesResponse,
    LinkedInLikes,
    LinkedInSocialDetail,
    LinkedInText,
    LinkedInCommentary,
    LinkedInProfilePostElement,
    LinkedInProfilePostsResponse,
    LinkedInProfileSkillsResponse,
    LinkedInCommentElement,
    LinkedInPostCommentResponse,
    LinkedInVoyagerValueV2,
    LinkedInUpdateValue,
    LinkedInUpdateElement,
    LinkedInUpdatesResponse,
    LinkedInSelfProfile,
    LinkedInJobSearchElement,
    LinkedInJobSearchResponse,
    OffsiteApply,
    ComplexApply,
    LinkedInApplyMethod,
    LinkedInJob,
    SkillMatchStatus,
    LinkedInJobSkills,
)
from .utils.query_options import (
    CompanyID,
    Experience,
    NetworkDepth,
    JobType,
    SortOrder,
    SortBy,
    JobTitle,
    GeoID,
    LocationType,
)

__title__ = "li_scrapi"
__version__ = "1.0.0"
__description__ = "Scraping API for LinkedIn, Built on the back of linkedin_api by Tom Quirk"

__license__ = "Apache License, Version 2.0"

__author__ = "Khalid Al-Awady"
__email__ = "pipes_and_bottles@protonmail.com"

__all__ = [
    "LinkedIn",
    "LinkedInScriptApi",
    "AsyncLinkedInClient",
    "LinkedInClient",
    "AsyncLinkedIn",
    "ContentSource",
    "JobState",
    "ContentType",
    "DistanceValue",
    "EnabledIndicators",
    "VisibilitySettings",
    "MiniProfile",
    "LinkedInExperience",
    "LinkedInEducation",
    "LinkedInPublication",
    "LinkedInCertification",
    "LinkedInVolunteer",
    "LinkedInProject",
    "LinkedInProfile",
    "LinkedInPrivacySettings",
    "LinkedInMemberBadges",
    "LinkedInNetwork",
    "CompanyIndustry",
    "OfficeLocation",
    "LinkedInOrganization",
    "LinkedInOrganizationResponse",
    "TwitterContactInfo",
    "WebsiteContactInfo",
    "LinkedInContactInfo",
    "LinkedInSearchElement",
    "LinkedInSearchPeopleElement",
    "LinkedInSearchCompaniesElement",
    "LinkedInSearchPeopleResponse",
    "LinkedInSearchCompaniesResponse",
    "LinkedInLikes",
    "LinkedInSocialDetail",
    "LinkedInText",
    "LinkedInCommentary",
    "LinkedInProfilePostElement",
    "LinkedInProfilePostsResponse",
    "LinkedInProfileSkillsResponse",
    "LinkedInCommentElement",
    "LinkedInPostCommentResponse",
    "LinkedInVoyagerValueV2",
    "LinkedInUpdateValue",
    "LinkedInUpdateElement",
    "LinkedInUpdatesResponse",
    "LinkedInSelfProfile",
    "LinkedInJobSearchElement",
    "LinkedInJobSearchResponse",
    "OffsiteApply",
    "ComplexApply",
    "LinkedInApplyMethod",
    "LinkedInJob",
    "SkillMatchStatus",
    "LinkedInJobSkills",
    "Experience",
    "NetworkDepth",
    "JobType",
    "LocationType",
    "SortOrder",
    "SortBy",
    "CompanyID",
    "JobTitle",
    "GeoID",
]



"""
MIT License

Copyright (c) 2024 Tom Quirk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

import http
from typing import Optional
from httpx import Cookies

from li_scrapi.utils import query_options
from li_scrapi.utils._base_linkedin import BaseLinkedIn
from li_scrapi.utils.schemas import (
    LinkedInProfile,
    LinkedInPrivacySettings,
    LinkedInMemberBadges,
    LinkedInNetwork,
    LinkedInOrganizationResponse,
    LinkedInContactInfo,
    LinkedInSearchPeopleResponse,
    LinkedInProfilePostsResponse,
    LinkedInProfileSkillsResponse,
    LinkedInPostCommentResponse,
    LinkedInSearchCompaniesResponse,
    LinkedInUpdatesResponse,
    LinkedInSelfProfile,
    LinkedInJobSearchResponse,
    LinkedInJob,
    LinkedInJobSkills,
)


class LinkedIn(BaseLinkedIn):
    """
    Class for accessing the LinkedIn API.

    :param username: Username of LinkedIn account.
    :type username: str
    :param password: Password of LinkedIn account.
    :type password: str
    """

    def _fetch(self, uri: str, base_request=False, **kwargs):
        """GET request to Linkedin API"""

        url = f"{self.client.API_BASE_URL if not base_request else self.client.LINKEDIN_BASE_URL}{uri}"
        return self.client.get(url, **kwargs)

    def _post(self, uri: str, base_request=False, **kwargs):
        """POST request to Linkedin API"""

        url = f"{self.client.API_BASE_URL if not base_request else self.client.LINKEDIN_BASE_URL}{uri}"
        return self.client.post(url, **kwargs)

    def _close(self):
        self.client.close()

    def authenticate(
        self, username: str, password: str, cookies: Optional[Cookies] = None
    ):
        if not cookies or not self.client._cookie_repository._is_token_still_valid(
            cookies.jar
        ):
            self.client.authenticate(username, password)
        else:
            self.client._set_session_cookies(cookies)
            self.client._cookie_repository.save(cookies, username)

    def get_profile_posts(
        self, urn_id: str, post_count=10, start=0, pagination_token=""
    ) -> LinkedInProfilePostsResponse | None:
        url, url_params = self._api_builder.build_profile_posts_url_params(
            urn_id,
            count=min(post_count, self._MAX_POST_COUNT),
            start=start,
            pagination_token=pagination_token,
        )
        linkedin_response = self._fetch(url, params=url_params)
        data = linkedin_response.json()
        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get profile posts data, response: {linkedin_response.status_code}"
            )
            return None
        return LinkedInProfilePostsResponse(**data)

    def get_post_comments(
        self,
        social_detail_urn: str,
        sort_by: query_options.SortOrder = query_options.SortOrder.RELEVANCE,
        comment_count=100,
        start=0,
        pagination_token="",
    ) -> LinkedInPostCommentResponse | None:
        url, url_params = self._api_builder.build_post_comments_url_query(
            post_urn=social_detail_urn,
            count=min(comment_count, self._MAX_POST_COUNT),
            start=start,
            sort_order=sort_by.value,
            pagination_token=pagination_token,
        )
        linkedin_response = self._fetch(url, params=url_params)
        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get post comments data, response: {linkedin_response.status_code}"
            )
            return None
        data = linkedin_response.json()

        return LinkedInPostCommentResponse(**data)

    def search(self, params: dict, offset=0) -> dict:
        url, _ = self._api_builder.build_search_url_params_query(params, offset)

        linkedin_response = self._fetch(url)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get search data, response: {linkedin_response.status_code}"
            )
            return {}

        data = linkedin_response.json()

        paging = self._parse_search_data(data)

        self.logger.debug(f"results grew to {len(paging.get("elements", []))}")

        return paging

    def search_people(
        self,
        keywords: str = "",
        connection_of: str = "",
        network_depths: list[query_options.NetworkDepth] = [],
        current_company: list[query_options.CompanyID] = [],
        past_companies: list[query_options.CompanyID] = [],
        nonprofit_interests: list[str] = [],
        profile_languages: list[str] = [],
        regions: list[query_options.GeoID] = [],
        industries: list[str] = [],
        schools: list[str] = [],
        service_categories: list[str] = [],
        include_private_profiles=False,  # profiles without a public id, "Linkedin Member"
        keyword_first_name: str = "",
        keyword_last_name: str = "",
        keyword_title: str = "",
        keyword_company: str = "",
        keyword_school: str = "",
        offset: int = 0,
    ) -> LinkedInSearchPeopleResponse:
        kwargs = {
            "connection_of": connection_of,
            "network_depths": [depth.value for depth in network_depths],
            "regions": [f"{region.value}" for region in regions],
            "industries": industries,
            "current_company": [f"{company.value}" for company in current_company],
            "past_companies": [f"{company.value}" for company in past_companies],
            "profile_languages": profile_languages,
            "nonprofit_interests": nonprofit_interests,
            "schools": schools,
            "service_categories": service_categories,
            "keyword_first_name": keyword_first_name,
            "keyword_last_name": keyword_last_name,
            "keyword_title": keyword_title,
            "keyword_company": keyword_company,
            "keyword_school": keyword_school,
        }

        params = self._api_builder.build_search_people_params(**kwargs)

        if keywords:
            params["keywords"] = keywords

        data = self.search(params, offset=offset)

        results = self._normalize_search_people_data(data, include_private_profiles)
        data["elements"] = results
        return LinkedInSearchPeopleResponse(**data)

    def search_companies(
        self, keywords: str = "", offset: int = 0
    ) -> LinkedInSearchCompaniesResponse:
        params = self._api_builder.build_search_companies_params(keywords)

        data = self.search(params, offset=offset)

        results = self._normalize_search_company_data(data)
        data["elements"] = results

        return LinkedInSearchCompaniesResponse(**data)

    def search_jobs(
        self,
        keywords: str = "",
        companies: list[query_options.CompanyID] = [],
        experience: list[query_options.Experience] = [],
        job_type: list[query_options.JobType] = [],
        job_title: list[query_options.JobTitle] = [],
        industries: list[str] = [],
        location: query_options.GeoID | None = None,
        remote: list[query_options.LocationType] = [],
        listed_at=24 * 60 * 60,
        distance: int | None = None,
        sort_by: query_options.SortBy = query_options.SortBy.RELEVANCE,
        v2: bool = False,
        limit=10,
        offset=0,
    ) -> LinkedInJobSearchResponse | None:
        count = min(limit, self._MAX_SEARCH_COUNT)

        kwargs = {
            "keywords": keywords,
            "experience": [f"{exp.value}" for exp in experience],
            "jobType": [job.value for job in job_type],
            "workplaceType": [f"{rem.value}" for rem in remote],
            "locationUnion": f"{location.value}" if location else "",
            "company": [f"{company.value}" for company in companies],
            "title": [f"{title.value}" for title in job_title],
            "industry": industries,
            "distance": distance,
            "sortBy": sort_by.value,
            "timePostedRange": listed_at,
        }

        url, _ = self._api_builder.build_search_jobs_url_query(count, offset, **kwargs)

        headers = {}
        if v2:
            headers = self._api_builder.get_v2_headers()

        linkedin_response = self._fetch(url, headers=headers, timeout=10.0)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get jobs, response: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()

        paging = self._normalize_search_jobs_data(data, v2)

        self.logger.debug(f"Collected {paging["paging"]["count"]} jobs")

        return LinkedInJobSearchResponse(**paging)

    def get_profile_contact_info(
        self, public_id: str = "", urn_id: str = ""
    ) -> LinkedInContactInfo | None:
        if not public_id and not urn_id:
            self.logger.info("Need to specifiy either a public id or a urn id")
            return None

        url = self._api_builder.build_contact_info_url(public_id, urn_id)
        res = self._fetch(url)
        data = res.json()

        contact_info = self._normalize_contact_info(data)

        return LinkedInContactInfo(**contact_info)

    def get_profile_skills(
        self, public_id: str = "", urn_id: str = ""
    ) -> LinkedInProfileSkillsResponse | None:
        if not public_id and not urn_id:
            self.logger.info("Need either public id or urn id to continue")
            return None

        params = self._api_builder.get_default_params()
        url = self._api_builder.build_skills_info_url(public_id, urn_id)
        linkedin_response = self._fetch(url, params=params)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get profile skills, response: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()

        return LinkedInProfileSkillsResponse(**data)

    def get_profile(
        self, public_id: str = "", urn_id: str = ""
    ) -> LinkedInProfile | None:
        if not public_id and not urn_id:
            self.logger.info("Need either public id or urn id to continue")
            return None

        url = self._api_builder.build_get_profile_url(public_id, urn_id)
        linkedin_response = self._fetch(url)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get profile data, response: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()

        # massage [profile] data
        profile = self._normalize_person_data(data)

        return LinkedInProfile(**profile)

    def get_profile_connections(
        self, urn_id: str, offset: int = 0
    ) -> LinkedInSearchPeopleResponse:
        return self.search_people(
            connection_of=urn_id,
            network_depths=[query_options.NetworkDepth.FIRST],
            offset=offset,
        )

    def get_company_updates(
        self,
        public_id: str = "",
        urn_id: str = "",
        start: int = 0,
        count: int = BaseLinkedIn._MAX_UPDATE_COUNT,
    ) -> LinkedInUpdatesResponse | None:
        if not public_id and not urn_id:
            self.logger.info("Need either public id or urn id to continue")
            return None

        url, params = self._api_builder.build_company_updates_url_params(
            public_id, urn_id, min(count, self._MAX_UPDATE_COUNT), start
        )

        linkedin_response = self._fetch(url, params=params)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"Failed to collect company data, error code: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()
        paging = data["paging"]
        paging["elements"] = data["elements"]

        self.logger.debug(f"results grew: {data["paging"]["count"]}")

        return LinkedInUpdatesResponse(paging=data["paging"], elements=data["elements"])

    def get_profile_updates(
        self,
        public_id: str = "",
        urn_id: str = "",
        start: int = 0,
        count: int = BaseLinkedIn._MAX_UPDATE_COUNT,
    ):
        if not public_id and not urn_id:
            self.logger.info("Need either public id or urn id to continue")
            return None

        url, params = self._api_builder.build_profile_updates_url_params(
            public_id, urn_id, min(count, self._MAX_UPDATE_COUNT), start
        )

        linkedin_response = self._fetch(url, params=params)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get profile updates, response: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()
        return LinkedInUpdatesResponse(**data)

    def get_organization(self, public_id: str):
        url, params = self._api_builder.build_get_organization_url_params(public_id)

        linkedin_response = self._fetch(url, params=params)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get organization data, response: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()
        paging = data["paging"]
        elements = data["elements"][0]

        return LinkedInOrganizationResponse(paging=paging, elements=elements)

    def get_user_profile(self, use_cache=True) -> LinkedInSelfProfile | None:
        me_profile = self._metadata.get("me")
        if not me_profile and (not self.client.metadata.get("me") or not use_cache):
            linkedin_response = self._fetch("/me")

            if linkedin_response.status_code != http.HTTPStatus.OK:
                self.logger.warning(
                    f"failed to get user profile data, response: {linkedin_response.status_code}"
                )
                return me_profile

            me_profile = LinkedInSelfProfile(**linkedin_response.json())
            # cache profile
            self._metadata["me"] = me_profile

        return me_profile

    def get_profile_privacy_settings(self, public_profile_id: str):
        res = self._fetch(
            self._api_builder.build_privacy_settings_url(public_profile_id),
            headers=self._api_builder.get_v2_headers(),
        )
        if res.status_code != http.HTTPStatus.OK:
            return None

        data = res.json()
        return LinkedInPrivacySettings(**data.get("data", {}))

    def get_profile_member_badges(self, public_profile_id: str):
        res = self._fetch(
            self._api_builder.build_member_badges_url(public_profile_id),
            headers=self._api_builder.get_v2_headers(),
        )
        if res.status_code != http.HTTPStatus.OK:
            return None

        data = res.json()
        return LinkedInMemberBadges(**data.get("data", {}))

    def get_profile_network_info(self, public_profile_id: str):
        res = self._fetch(
            self._api_builder.build_profile_network_url(public_profile_id),
            headers=self._api_builder.get_v2_headers(),
        )
        if res.status_code != http.HTTPStatus.OK:
            return None

        data = res.json()
        return LinkedInNetwork(**data.get("data", {}))

    def get_job(self, job_id: str) -> LinkedInJob | None:
        url, params = self._api_builder.build_get_job_url_params(job_id)

        linkedin_response = self._fetch(url, params=params)

        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get job data, response: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()
        data["workplaceTypes"] = [
            query_options.LocationType(int(val.split(":")[-1]))
            for val in data["workplaceTypes"]
        ]

        return LinkedInJob(**data)

    def get_job_skills(self, job_id: str) -> LinkedInJobSkills | None:
        url, params = self._api_builder.build_get_job_skill_url_params(job_id)
        linkedin_response = self._fetch(
            url,
            params=params,
        )
        if linkedin_response.status_code != http.HTTPStatus.OK:
            self.logger.warning(
                f"failed to get job data, response: {linkedin_response.status_code}"
            )
            return None

        data = linkedin_response.json()

        return LinkedInJobSkills(**data)



"""
MIT License

Copyright (c) 2024 Tom Quirk

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import loguru
from abc import ABC, abstractmethod
from httpx import Cookies
from li_scrapi.client import Client
from operator import itemgetter
from li_scrapi.utils.helpers import (
    get_id_from_urn,
    get_urn_from_raw_update,
)
from li_scrapi.utils.api_builder import ApiBuilder
from li_scrapi.utils import query_options
from li_scrapi.utils.schemas import (
    LinkedInProfile,
    LinkedInPrivacySettings,
    LinkedInMemberBadges,
    LinkedInNetwork,
    LinkedInOrganizationResponse,
    LinkedInContactInfo,
    LinkedInSearchPeopleResponse,
    LinkedInProfilePostsResponse,
    LinkedInProfileSkillsResponse,
    LinkedInPostCommentResponse,
    LinkedInSearchCompaniesResponse,
    LinkedInUpdatesResponse,
    LinkedInSelfProfile,
    LinkedInJobSearchResponse,
    LinkedInJob,
    LinkedInJobSkills,
)


class BaseLinkedIn(ABC):
    _MAX_POST_COUNT = 100  # max seems to be 100 posts per page
    _MAX_UPDATE_COUNT = 100  # max seems to be 100
    _MAX_SEARCH_COUNT = 49  # max seems to be 49, and min seems to be 2
    _MAX_REPEATED_REQUESTS = (
        200  # VERY conservative max requests count to avoid rate-limit
    )

    logger = loguru.logger
    _api_builder = ApiBuilder()

    def __init__(
        self,
        client: Client,
    ):
        """Constructor method"""
        self.client = client
        self._metadata: dict[str, LinkedInSelfProfile] = {}

    def _cookies(self):
        """Return client cookies"""
        return self.client.cookies

    def _headers(self):
        """Return client cookies"""
        return self.client.REQUEST_HEADERS

    def _normalize_person_data(self, data: dict) -> dict:
        profile = data["profile"]
        if "miniProfile" in profile:
            if "picture" in profile["miniProfile"]:
                profile["displayPictureUrl"] = profile["miniProfile"]["picture"][
                    "com.linkedin.common.VectorImage"
                ]["rootUrl"]

                images_data = profile["miniProfile"]["picture"][
                    "com.linkedin.common.VectorImage"
                ]["artifacts"]
                for img in images_data:
                    w, h, url_segment = itemgetter(
                        "width", "height", "fileIdentifyingUrlPathSegment"
                    )(img)
                    profile[f"img_{w}_{h}"] = url_segment

            profile["urn_id"] = get_id_from_urn(profile["miniProfile"]["entityUrn"])
            profile["profile_urn"] = profile["miniProfile"]["entityUrn"]
            profile["member_urn"] = profile["miniProfile"]["objectUrn"]
            profile["public_id"] = profile["miniProfile"]["publicIdentifier"]

        # massage [experience] data
        experience = data["positionView"]["elements"]
        for item in experience:
            if "company" in item and "miniCompany" in item["company"]:
                if "logo" in item["company"]["miniCompany"]:
                    logo = item["company"]["miniCompany"]["logo"].get(
                        "com.linkedin.common.VectorImage"
                    )
                    if logo:
                        item["companyLogoUrl"] = logo["rootUrl"]
                del item["company"]["miniCompany"]

        profile["experience"] = experience

        # massage [education] data
        education = data["educationView"]["elements"]
        for item in education:
            if "school" in item:
                if "logo" in item["school"]:
                    item["school"]["logoUrl"] = item["school"]["logo"][
                        "com.linkedin.common.VectorImage"
                    ]["rootUrl"]
                    del item["school"]["logo"]

        profile["education"] = education

        # massage [languages] data
        profile["languages"] = data["languageView"]["elements"]

        # massage [publications] data
        profile["publications"] = data["publicationView"]["elements"]

        # massage [certifications] data
        profile["certifications"] = data["certificationView"]["elements"]

        # massage [volunteer] data
        profile["volunteer"] = data["volunteerExperienceView"]["elements"]

        # massage [honors] data
        profile["honors"] = data["honorView"]["elements"]

        # massage [projects] data
        profile["projects"] = data["projectView"]["elements"]

        # massage [skills] data
        profile["skills"] = data["skillView"]["elements"]

        return profile

    def _parse_search_data(self, data: dict) -> dict:
        data_clusters: dict = data.get("data", {}).get("searchDashClustersByAll", {})

        if not data_clusters:
            return {}

        if (
            data_clusters.get("_type", "")
            != "com.linkedin.restli.common.CollectionResponse"
        ):
            return {}

        new_elements = []
        for it in data_clusters.get("elements", []):
            if (
                it.get("_type", "")
                != "com.linkedin.voyager.dash.search.SearchClusterViewModel"
            ):
                continue

            for el in it.get("items", []):
                if el.get("_type", "") != "com.linkedin.voyager.dash.search.SearchItem":
                    continue

                e = el.get("item", {}).get("entityResult", [])
                if not e:
                    continue
                if (
                    e.get("_type", "")
                    != "com.linkedin.voyager.dash.search.EntityResultViewModel"
                ):
                    continue
                new_elements.append(e)
        result = {"paging": data_clusters["paging"], "elements": new_elements}
        return result

    def _normalize_search_people_data(
        self, data: dict, include_private_profiles: bool
    ) -> list[dict]:
        results = []
        for item in data.get("elements", []):
            if (
                not include_private_profiles
                and (item.get("entityCustomTrackingInfo") or {}).get(
                    "memberDistance", None
                )
                == "OUT_OF_NETWORK"
            ):
                continue
            results.append(
                {
                    "urn_id": get_id_from_urn(
                        get_urn_from_raw_update(item.get("entityUrn", None))
                    ),
                    "distance": (item.get("entityCustomTrackingInfo") or {}).get(
                        "memberDistance", None
                    ),
                    "job_title": (item.get("primarySubtitle") or {}).get("text", None),
                    "location": (item.get("secondarySubtitle") or {}).get("text", None),
                    "name": (item.get("title") or {}).get("text", None),
                }
            )
        return results

    def _normalize_search_jobs_data(self, data: dict, v2: bool) -> dict:
        results = []
        if v2:
            elements = data.get("included", [])
            for element in elements:
                if element["$type"] == "com.linkedin.voyager.dash.jobs.JobPosting":
                    results.append(element)
        else:
            elements = data.get("elements", [])
            for element in elements:
                job_data = (
                    element.get("jobCardUnion", {})
                    .get("jobPostingCard", {})
                    .get("jobPosting", {})
                )
                if job_data:
                    results.append(job_data)

        if v2:
            paging = data["data"]["paging"]
        else:
            paging = data.get("paging", {})

        return {"paging": paging, "elements": results}

    def _normalize_search_company_data(self, data: dict) -> list[dict]:
        results = []
        for item in data.get("elements", []):
            if "company" not in item.get("trackingUrn"):
                continue
            results.append(
                {
                    "urn_id": get_id_from_urn(item.get("trackingUrn", None)),
                    "name": (item.get("title") or {}).get("text", None),
                    "headline": (item.get("primarySubtitle") or {}).get("text", None),
                    "subline": (item.get("secondarySubtitle") or {}).get("text", None),
                }
            )
        return results

    def _normalize_contact_info(self, data: dict) -> dict:
        contact_info = {
            "email_address": data.get("emailAddress"),
            "websites": [],
            "twitter": data.get("twitterHandles"),
            "birthdate": data.get("birthDateOn"),
            "ims": data.get("ims"),
            "phone_numbers": data.get("phoneNumbers", []),
        }

        websites = data.get("websites", [])
        for item in websites:
            if "com.linkedin.voyager.identity.profile.StandardWebsite" in item["type"]:
                item["label"] = item["type"][
                    "com.linkedin.voyager.identity.profile.StandardWebsite"
                ]["category"]
            elif "" in item["type"]:
                item["label"] = item["type"][
                    "com.linkedin.voyager.identity.profile.CustomWebsite"
                ]["label"]

            del item["type"]

        contact_info["websites"] = websites
        return contact_info

    @abstractmethod
    def _fetch(self, uri: str, base_request=False, **kwargs): ...

    @abstractmethod
    def _post(self, uri: str, base_request=False, **kwargs): ...

    @abstractmethod
    def authenticate(
        self, username: str, password: str, cookies: Cookies | None = None
    ): ...

    @abstractmethod
    def get_profile_posts(
        self, urn_id: str, post_count=10, start=0, pagination_token=""
    ) -> LinkedInProfilePostsResponse | None:
        """
        :param urn_id: LinkedIn URN ID for a profile
        :param post_count: Number of posts to fetch
        :param start: Pagination start for the posts to fetch
        :param pagination_token: Pagination token if more than 100 posts

        :return: LinkedInProfilePostsResponse | None
        """

    @abstractmethod
    def get_post_comments(
        self,
        social_detail_urn: str,
        sort_by: query_options.SortOrder = query_options.SortOrder.RELEVANCE,
        comment_count=100,
        start=0,
        pagination_token="",
    ) -> LinkedInPostCommentResponse | None:
        """
        get_post_comments: Get post comments

        :param social_detail_urn: Post URN
        :param sort_by: sort param for the fetch request
        :param comment_count: Number of comments to fetch
        :param start: offset for pagination on fetch
        :param pagination_token: Pagination token if more than 100 posts

        :return: LinkedInPostCommentResponse | None
        """

    @abstractmethod
    def search(self, params: dict, offset=0) -> dict:
        """
        Perform a LinkedIn search. This function should never
        be used unless you know what you are doing since the JSONs are too nested

        :param params: Search parameters (see other search methods for an idea of what to pass)
        :param offset: Index to start searching from

        :return: dictionary
        """

    @abstractmethod
    def search_people(
        self,
        keywords: str = "",
        connection_of: str = "",
        network_depths: list[query_options.NetworkDepth] = [],
        current_company: list[query_options.CompanyID] = [],
        past_companies: list[query_options.CompanyID] = [],
        nonprofit_interests: list[str] = [],
        profile_languages: list[str] = [],
        regions: list[query_options.GeoID] = [],
        industries: list[str] = [],
        schools: list[str] = [],
        service_categories: list[str] = [],
        include_private_profiles=False,  # profiles without a public id, "Linkedin Member"
        keyword_first_name: str = "",
        keyword_last_name: str = "",
        keyword_title: str = "",
        keyword_company: str = "",
        keyword_school: str = "",
        offset: int = 0,
    ) -> LinkedInSearchPeopleResponse:
        """Perform a LinkedIn search for people.

        :param keywords: Keywords to search on
        :param connection_of: Connection of LinkedIn user, given by profile URN ID
        :param network_depths: A list containing one or many of "F", "S" and "O" (first, second and third+ respectively)
        :param current_company: A list of company URN IDs
        :param past_companies: A list of company URN IDs
        :param nonprofit_interests: A list of non-profit URN ID (?)
        :param profile_languages: A list of 2-letter language codes
        :param regions: A list of geo URN IDs
        :param industries: A list of industry URN IDs
        :param schools: A list of school URN IDs
        :param service_categories: A list of service category URN IDs
        :param include_private_profiles: Include private profiles in search results. If False, only public profiles are included
        :param keyword_first_name: First name
        :param keyword_last_name: Last name
        :param keyword_title: Job title
        :param keyword_company: Company name
        :param keyword_school: School name
        :param offset: pagination index of search

        :return: LinkedInSearchPeopleResponse
        """

    @abstractmethod
    def search_companies(
        self, keywords: str = "", offset: int = 0
    ) -> LinkedInSearchCompaniesResponse:
        """
        Perform a LinkedIn search for companies.

        :param keywords: key words to look up, optional
        :param offset: offset for pagination in search

        :return: LinkedInSearchCompaniesResponse
        """

    @abstractmethod
    def search_jobs(
        self,
        keywords: str = "",
        companies: list[query_options.CompanyID] = [],
        experience: list[query_options.Experience] = [],
        job_type: list[query_options.JobType] = [],
        job_title: list[query_options.JobTitle] = [],
        industries: list[str] = [],
        location: query_options.GeoID | None = None,
        remote: list[query_options.LocationType] = [],
        listed_at=24 * 60 * 60,
        distance: int | None = None,
        sort_by: query_options.SortBy = query_options.SortBy.RELEVANCE,
        v2: bool = False,
        limit=10,
        offset=0,
    ) -> LinkedInJobSearchResponse | None:
        """
        Perform a LinkedIn search for jobs.

        :param keywords: Search keywords
        :param companies: A list of company URN IDs
        :param experience: A list of experience levels, one or many of (internship, entry level, associate, mid-senior level, director and executive, respectively)
        :param job_type:  A list of job types , one or many of (full-time, contract, part-time, temporary, internship, volunteer and "other", respectively)
        :param job_title: A list of title URN IDs
        :param industries: A list of industry URN IDs
        :param location: Name of the location to search within. Example: "Kyiv City, Ukraine"
        :param remote: Filter for remote jobs, onsite or hybrid. onsite:"1", remote:"2", hybrid:"3"
        :param listed_at: maximum number of seconds passed since job posting. 86400 will filter job postings posted in last 24 hours.
        :param distance: maximum distance from location in miles
        :param sort_by: Filter for how to sort jobs, by one of relevancy or time posted
        :param limit: maximum number of results obtained from API queries. -1 means maximum which is defined by constants and is equal to 1000 now.
        :param offset: indicates how many search results shall be skipped

        :return: LinkedInJobSearchResponse | None
        """

    @abstractmethod
    def get_profile_contact_info(
        self, public_id: str = "", urn_id: str = ""
    ) -> LinkedInContactInfo | None:
        """
        Fetch contact information for a given LinkedIn profile. Pass a [public_id] or a [urn_id].

        :param public_id: LinkedIn public ID for a profile
        :param urn_id: LinkedIn URN ID for a profile

        :return: LinkedInContactInfo | None
        """

    @abstractmethod
    def get_profile_skills(
        self, public_id: str = "", urn_id: str = ""
    ) -> LinkedInProfileSkillsResponse | None:
        """
        Fetch the skills listed on a given LinkedIn profile.

        :param public_id: LinkedIn public ID for a profile
        :param urn_id: LinkedIn URN ID for a profile

        :return:  LinkedInProfileSkillsResponse | None
        """

    @abstractmethod
    def get_profile(
        self, public_id: str = "", urn_id: str = ""
    ) -> LinkedInProfile | None:
        """
        Fetch the skills listed on a given LinkedIn profile.

        :param public_id: LinkedIn public ID for a profile
        :param urn_id: LinkedIn URN ID for a profile

        :return:LinkedInProfile | None
        """

    @abstractmethod
    def get_profile_connections(
        self, urn_id: str, offset: int = 0
    ) -> LinkedInSearchPeopleResponse:
        """
        Fetch first-degree connections for a given LinkedIn profile.

        :param urn_id: LinkedIn URN ID for a profile
        :param offset: indicates how many search results shall be skipped

        :return: LinkedInSearchPeopleResponse
        """

    @abstractmethod
    def get_company_updates(
        self,
        public_id: str = "",
        urn_id: str = "",
        start: int = 0,
        count: int = _MAX_UPDATE_COUNT,
    ) -> LinkedInUpdatesResponse | None:
        """
        Fetch company updates (news activity) for a given LinkedIn company.

        :param public_id: LinkedIn public ID for a company
        :param urn_id: LinkedIn URN ID for a company
        :param start: offset for query due to pagination
        :param count: number of items to pull from LinkedIn

        :return: LinkedInUpdatesResponse | None
        """

    @abstractmethod
    def get_profile_updates(
        self,
        public_id: str = "",
        urn_id: str = "",
        start: int = 0,
        count: int = _MAX_UPDATE_COUNT,
    ) -> LinkedInUpdatesResponse | None:
        """
        Fetch profile updates (newsfeed activity) for a given LinkedIn profile.

        :param public_id: LinkedIn public ID for a profile
        :param urn_id: LinkedIn URN ID for a profile
        :param start: offset for query due to pagination
        :param count: number of items to pull from LinkedIn

        :return: LinkedInUpdatesResponse | None
        """

    @abstractmethod
    def get_organization(self, public_id: str) -> LinkedInOrganizationResponse | None:
        """
        Fetch data about a given school or Company on LinkedIn.

        :param public_id: LinkedIn public ID for a school
        :return: LinkedInOrganizationResponse | None
        """

    @abstractmethod
    def get_user_profile(self, use_cache=True) -> LinkedInSelfProfile | None:
        """
        Get the current user profile. If not cached, a network request will be fired.

        :return: LinkedInSelfProfile | None
        """

    @abstractmethod
    def get_profile_privacy_settings(
        self, public_profile_id: str
    ) -> LinkedInPrivacySettings | None:
        """
        Fetch privacy settings for a given LinkedIn profile.

        :param public_profile_id: public ID of a LinkedIn profile

        :return: LinkedInPrivacySettings | None
        """

    @abstractmethod
    def get_profile_member_badges(
        self, public_profile_id: str
    ) -> LinkedInMemberBadges | None:
        """
        Fetch badges for a given LinkedIn profile.

        :param public_profile_id: public ID of a LinkedIn profile

        :return: LinkedInMemberBadges | None
        """

    @abstractmethod
    def get_profile_network_info(
        self, public_profile_id: str
    ) -> LinkedInNetwork | None:
        """
        Fetch network information for a given LinkedIn profile.

        :param public_profile_id: public ID of a LinkedIn profile

        :return: LinkedInNetwork | None
        """

    @abstractmethod
    def get_job(self, job_id: str) -> LinkedInJob | None:
        """
        Fetch data about a given job.

        :param job_id: LinkedIn job ID

        :return: LinkedInJob | None
        """

    @abstractmethod
    def get_job_skills(self, job_id: str) -> LinkedInJobSkills | None:
        """
        Fetch skills associated with a given job.

        :param job_id: LinkedIn job ID

        :return: LinkedInJobSkills | None
        """
